/*------------------------------------------------------------------*/
/* File: C:\cbproj\YHF_Tools\YHF_Inet.c                             */
/*       (everything else is just a COPY)                           */
/*                                                                  */
/* DL4YHF's 'internet' utilities based on the socket services .     */
/* Initially used to read audio streams from remote servers .       */
/* Later extended to support a few of those bazillions of encodings */
/* (like "Base64") and checksums / hashes (like "SHA1") required    */
/* for various web services, from HTTP to "WebSockets".             */
/*                                                                  */
/*                                                                  */
/* Autor: Wolfgang Buescher (DL4YHF), 2011 ... 2019                 */
/*  Used in:                                                        */
/*    *  Spectrum Lab, to retrieve the contents of a remote *.M3U,  */
/*    *  in SL's "OggVorbis"-DLL, to read Ogg/Vorbis streams,       */
/*    *  Spectrum Lab, to "serve" HTTP, "WebAudio" and "WebSockets",*/
/*    *  and whatever comes next, to make this even more bizarre .  */
/*                                                                  */
/* Revision history (YYYY-MM-DD):                                   */
/*                                                                  */
/*  2019-01-20 :                                                    */
/*   Added a few utility functions required for WebSockets, like... */
/*      Base64 encoding and decoding (for the "Sec-WebSocket-Key"), */
/*      SHA-1 calculation (also for the WebSocket-key), etc .       */
/*                                                                  */
/*  2012-04-26 :                                                    */
/*   Modified the transmit-side (INET_Write) to defeat              */
/*      temporary bandwidth problems on Renato's VLF-streaming PC.  */
/*      YHF_Inet.c now uses a much larger buffer for *OUTGOING*     */
/*      traffic (up to 512 kByte!) and should be used with the      */
/*      option 'INET_WR_OPTION_DONT_FLUSH' wherever possible .      */
/*                                                                  */
/*  2011-12-26 :                                                    */
/*   Changed from the old (simple, but inefficient) polling-method  */
/*      to asynchronous sockets (with a window message sent when    */
/*      new data have arrived. NO MULTI-THREADING) . Details at     */
/*      http://tangentsoft.net/wskfaq/articles/io-strategies.html . */
/*      Old version saved as C:\cbproj\YHF_Tools\YHF_Inet_simple.c  */
/*                                                                  */
/*  2011-11-30 :                                                    */
/*  Created YHF_Inet.c as an extra layer between an application     */
/*      (CLIENT) to read files from remote internet servers.        */
/*      Contains an afwul lot of specialized 'string parsers'       */
/*      to separate URLs (or do they call it URIs these days?),     */
/*      for example, to ...                                         */
/*   - find out if a given "filename" is the name of a LOCAL FILE   */
/*      or an INTERNET RESOURCE (to pick the right API to access    */
/*      them, because unfortunately, under windoze, a SOCKET        */
/*      is a SOCKET and not a FILE HANDLE, and vice versa... )      */
/*   - parse strings like "http://67.207.143.181/vlf6.m3u"  ,       */
/*      to retrieve the PROTOCOL,  HOST NAME,  FILENAME ;           */
/*   - Simplify the task of opening a TCP/IP connection, and        */
/*      (on top of that) implement a simple HTTP CLIENT .           */
/*                                                                  */
/*------------------------------------------------------------------*/


// ex: #include "SWITCHES.H"  // project specific compiler switches ("options")
                       // must be included before anything else !



/* NO VCL IN THIS UNIT !!!!! (or other BORing stuff) */
#include <string.h>
#include <windows.h> // Can we eliminate this ? Nope, we need sockets and neither want POSIX nor bloated "libSomething" !
#include <stdarg.h>
#include <stdio.h>   // not really 'standard I/O' but vsprintf used here
#pragma hdrstop

#include "YHF_Inet.h"     // DL4YHF's internet helper functions
#include "YHF_CmdLine.h"  // DL4YHF's command-line-parser utility
#include "StringLib.h"    // DL4YHF's C-string processing library
#include "Utility1.h"     // UTL_CheckAndSkipToken() + similar parsing stuff
#include "Timers.h"       // high-resolution 'current time'-function, T_TIM_Stopwatch, etc
#include "ErrorCodes.h"   // DL4YHF's error codes (also used here since 2011)



//------------ internal defines ------------------------------------------

   // ... none so far ...



//------------ internal data types ------------------------------------------

//------------ Test data for some encoders and hash generators --------------
   // From de.wikipedia.org/wiki/Secure_Hash_Algorithm#Beispiel-Hashes :
static const char szFranz[] = "Franz jagt im komplett verwahrlosten Taxi quer durch Bayern";
static const char szFranzBase64[] =
   "RnJhbnogamFndCBpbSBrb21wbGV0dCB2ZXJ3YWhybG9zdGVuIFRheGkgcXVlciBkdXJjaCBCYXllcm4=";
static const BYTE b20FranzSHA1[] = { 0x68, 0xac, 0x90, 0x64,
       /* 160 bit / 8bit_per_byte*/  0x95, 0x48, 0x0a, 0x34,
                                     0x04, 0xbe, 0xee, 0x48,
                                     0x74, 0xed, 0x85, 0x3a,
                                     0x03, 0x7a, 0x7a, 0x8f };


//------------ Functions ----------------------------------------------------


//---------------------------------------------------------------------------
char * INET_WinsockErrorCodeToString( int iWSAErrorCode )
 // Turns error codes returned by WSAGetLastError() (!!!!) into text .
 //
{
  static char sz80OtherError[84];  // <<< avoid using this static var wherever possible
                                   //  (it may cause problems with multi-threading one day)
  switch( iWSAErrorCode )
   {
    case WSAEINTR          : return "Interrupted function call";
    case WSAEBADF          : return "Invalid socket descriptor";
    case WSAEACCES         : return "Permission denied";
    case WSAEFAULT         : return "Bad address";
    case WSAEINVAL         : return "Invalid argument";
    case WSAEMFILE         : return "Too many open files or sockets";
    case WSAEWOULDBLOCK    : return "Call of WINSOCK routine would block";
    case WSAEINPROGRESS    : return "Operation now in progress";
    case WSAEALREADY       : return "Operation already in progress";
    case WSAENOTSOCK       : return "Socket operation on non-socket";
    case WSAEDESTADDRREQ   : return "Destination address required";
    case WSAEMSGSIZE       : return "Message too long";
    case WSAEPROTOTYPE     : return "Protocol wrong type for socket";
    case WSAENOPROTOOPT    : return "Bad protocol option";
    case WSAEPROTONOSUPPORT: return "Socket type not supported";
    case WSAESOCKTNOSUPPORT: return "Socket not supported";
    case WSAEOPNOTSUPP     : return "Operation not supported";
    case WSAEPFNOSUPPORT   : return "Protocol family not supported";
    case WSAEAFNOSUPPORT   : return "Address family not supported by protocol family";
    case WSAEADDRINUSE     : return "Address already in use";
    case WSAEADDRNOTAVAIL  : return "Cannot assign requested address";
    case WSAENETDOWN       : return "Network is down";
    case WSAENETUNREACH    : return "Network is unreachable";
    case WSAENETRESET      : return "Network dropped connection on reset";
    case WSAECONNABORTED   : return "Software caused connection abort";
    case WSAECONNRESET     : return "Connection reset by peer";
    case WSAENOBUFS        : return "No buffer space available";
    case WSAEISCONN        : return "socket already connected";
    case WSAENOTCONN       : return "Socket is not connected";
    case WSAESHUTDOWN      : return "Cannot send after socket shutdown";
    case WSAETOOMANYREFS   : return "too many refs";
    case WSAETIMEDOUT      : return "Connection timed out";
    case WSAECONNREFUSED   : return "Connection refused";
    case WSAELOOP          : return "loop detected";
    case WSAENAMETOOLONG   : return "name too long";
    case WSAEHOSTDOWN      : return "Host is down";
    case WSAEHOSTUNREACH   : return "No route to host";
    case WSAENOTEMPTY      : return "not empty";
    case WSAEPROCLIM       : return "Too many processes";
    case WSAEUSERS         : return "users";
    case WSAEDQUOT         : return "quot";
    case WSAESTALE         : return "stale";
    case WSAEREMOTE        : return "remote";
    case WSAEDISCON        : return "Graceful shutdown in progress";
    case WSASYSNOTREADY    : return "Network subsystem is unavailable";
    case WSAVERNOTSUPPORTED: return "WINSOCK.DLL version out of range";
    case WSANOTINITIALISED : return "WSAStartup not yet performed";
    case WSAHOST_NOT_FOUND : return "Host not found";
    case WSATRY_AGAIN      : return "Non-authoritative host not found";
    case WSANO_RECOVERY    : return "non-recoverable error";
    case WSANO_DATA        : return "valid name but no data";
    default                :
         sprintf(sz80OtherError, "Other WINSOCK error, code %d",(int)iWSAErrorCode);
         return sz80OtherError;
   } // end switch( iWSAErrorCode )

} // end INET_WinsockErrorCodeToString()

/****************************************************************************/
int INET_GuessAudioFileTypeFromName(char * pszFilename,
                             int iDefaultAudioFileType) // [in] to solve "*.dat"-ambiguity !)
{
  char *cp = strrchr(pszFilename,'.');
  // char *cpFirstColon = strchr(pszFilename,':');
  if(cp!=NULL)
   {
     if(stricmp(cp+1,"raw")==0)
      { return AUDIO_FILE_FORMAT_RAW;
      }
     if(stricmp(cp+1,"dat")==0)        // .dat produced by AUDIO UTILITIES, also headerless raw data or STREAM LOGFILES
      { // In this case, there is an ambiguity between "old" raw files,
        // and the 8-bit I/Q format used by the SETI folks (also *.dat) :
        switch( iDefaultAudioFileType )
         { case AUDIO_FILE_FORMAT_SETI_1: // SETI 8-bit raw I/Q, *.dat, 8738133 samples/sec
              return iDefaultAudioFileType;
           case AUDIO_FILE_FORMAT_STREAM_WITH_HEADERS: // also preferrably logged as anonymous "*.dat" !
              return iDefaultAudioFileType;
           case AUDIO_FILE_FORMAT_STREAM_WITH_VLF_RX_BLOCKS :
              // Uncompressed stream from VLF-RX-tools *without* T_StreamHeaders
              //  but VT_BLOCKs.  Added here 2020-10-11 but not really supported yet.
              return iDefaultAudioFileType;
           default:  // cannot be MP3, WAW, etc !
              return AUDIO_FILE_FORMAT_RAW;
         }
      }
     if(stricmp(cp+1,"wav")==0)
      { return AUDIO_FILE_FORMAT_RIFF; // (2) standard "RIFF wave audio" (*.wav)
      }
     if( (stricmp(cp+1,"ogg")==0) || (stricmp(cp+1,"ogx")==0) || (stricmp(cp+1,"oga")==0) )
      { return AUDIO_FILE_FORMAT_OGG;  // Ogg(container) / Vorbis(codec) (*.ogg)
        //  .ogg, .spx = "audio/ogg" nach http://tools.ietf.org/html/rfc5334
        //  .ogx       = "application/ogg"
        //  .oga       = "audio/ogg" (im Gegensatz zu .ogg NUR Audio, kein Video)
      }
     if(stricmp(cp+1,"txt")==0)
      { return AUDIO_FILE_FORMAT_TEXT;
      }
 //  if(stricmp(cp+1,"jas")==0)
 //   { return AUDIO_FILE_FORMAT_JASON;
 //   }
     if(stricmp(cp+1,"m3u")==0)
      { return AUDIO_FILE_FORMAT_M3U; // not an AUDIO FILE FORMAT but a 'playlist' (text)
        // To detect the REAL type of this medium, the file (*.m3u) must be opened,
        // or loaded from a remote server !
        // Example:   http://67.207.143.181/vlf6.m3u   may contain something like this:
        //            http://67.207.143.181:80/vlf6    .
        // Opening that *STREAM* (through the socket services) turns up
        // the TRUE "file" format later (mp3 compression or ogg/vorbis) .
      }
     if(stricmp(cp+1,"mp3")==0)
      { return AUDIO_FILE_FORMAT_MP3; // not supported by SpecLab (requires plugin)
      }

     if(stricmp(cp+1,".vlfrx")==0)
      { // Uncompressed stream from VLF-RX-tools *without* T_StreamHeaders but VT_BLOCKs only:
        return AUDIO_FILE_FORMAT_STREAM_WITH_VLF_RX_BLOCKS;
        // Details in C:\cbproj\SoundUtl\VorbisFileIO.cpp (for historic reasons) !
      }

   } // end if < saw something which LOOKS LIKE a file extension >

  if( strnicmp( pszFilename, "http://", 7)== 0 ) // added 2011-12-05
   { if( iDefaultAudioFileType == AUDIO_FILE_FORMAT_UNKNOWN )
      { return AUDIO_FILE_FORMAT_OTHER_WEB_STREAM;
      }
     else
      { return iDefaultAudioFileType;
      }
   }

  if( strnicmp( pszFilename, "rawtcp://", 8)== 0 ) // added 2011-12-11 (dummy for "raw TCP", not HTTP)
   { if( iDefaultAudioFileType == AUDIO_FILE_FORMAT_UNKNOWN )
      { return AUDIO_FILE_FORMAT_OTHER_WEB_STREAM;
      }
     else
      { return iDefaultAudioFileType;
      }
   }

  if( strnicmp( pszFilename, "tcp://", 6)== 0 ) // added 2018-11 after seing "tcp://" in a config by DK7FC
   { if( iDefaultAudioFileType == AUDIO_FILE_FORMAT_UNKNOWN )
      { return AUDIO_FILE_FORMAT_OTHER_WEB_STREAM;
      }
     else
      { return iDefaultAudioFileType;
      }
   }


  return iDefaultAudioFileType;
} // end INET_GuessAudioFileTypeFromName()


/****************************************************************************/
int INET_GuessAudioFileTypeFromData(
       BYTE *pbFileData, int nBytes, // [in] a few hundred bytes read from the start of the FILE, or received from a web audio stream
       int iDefaultAudioFileType )   // [in] default type (AUDIO_FILE_FORMAT_..) when no match was found,
                                     //      or -in some cases- a "clue". Values defined in c:\cbproj\SoundUtl\AudioFileDefs.h
   // Usage : See C:\cbproj\SpecLab\SpecMain.cpp : TSpectrumLab::OpenInputFileAnalysis() .
   // See also : CheckBufferForUncompressedStreamHeader() and
   //            CheckBufferForVlfRxToolsBlock() in VorbisFileIO.cpp !
   //            (detects 'raw', uncompressed audio streams compatible with
   //             Paul Nicholson's VLFRXTOOLS )
{
  if( (pbFileData!=NULL) && (nBytes>=8) )
   { if( strncmp( (char*)pbFileData, "OggS", 4) == 0 )
      { return AUDIO_FILE_FORMAT_OGG;  // Ogg(container) / Vorbis(codec) (*.ogg)
      }

     // Similar as in CheckBufferForUncompressedStreamHeader()
     //           and CheckBufferForVlfRxToolsBlock(), but, to keep it simple,
     // here we assume a T_StreamHeader or a T_VT_Block is at the FIRST byte:
     if( (pbFileData[0] == 0xD3) && (pbFileData[1] == 0x6C) )
      { return AUDIO_FILE_FORMAT_STREAM_WITH_VLF_RX_BLOCKS; // it's a VT_BLOCK (alias T_VT_Block) !
        // Note that there's a chance of 1:65536 to get this WRONG
        // in a file with random data (or digitized white noise) !
      }

     // If a T_StreamHeader was available here (which it is not, for good reasons)
     // the test for a valid T_StreamHeader *pSH would be :
     // > if(  (pSH->dwPattern8000 == 0x80008000UL) // 1st criterion for a header structure
     if( (pbFileData[0]==0x00) && (pbFileData[1]==0x80)
      && (pbFileData[2]==0x00) && (pbFileData[3]==0x80) ) // T_StreamHeader.dwPattern8000, little endian ?
      { return AUDIO_FILE_FORMAT_STREAM_WITH_HEADERS; // yes, it's a stream identified by T_StreamHeader structs !
      }

   } // end if( (pbFileData!=NULL) && (nBytes>=8) )

  return iDefaultAudioFileType;
} // end INET_GuessAudioFileTypeFromData()


/****************************************************************************/
int INET_ContentTypeToAudioFileFormat( char *pszContentType, int iDefaultAudioFileType )
{
  BYTE *bp = (BYTE*)pszContentType;

  if( UTL_CheckAndSkipToken( &bp, "audio/x-mpegurl" ) )
   {  // This is what http://67.207.143.181/vlf6.m3u replied.
      // According to http://webdesign.about.com/od/sound/a/sound_mime_type.htm ,
      //  "audio/x-mpegurl" is only the MIME-type for "m3u" .
      // It doesn't tell us anything about the audio stream itself !
      //  (the stream itself may be OGG/VORBIS instead of MPEG)
      return AUDIO_FILE_FORMAT_M3U; // ok, it's definitely M3U
   }
  else if( UTL_CheckAndSkipToken( &bp, "audio/mpeg" ) )
   {  // Now we know this URL (pszAudioFileName) is not the name
      // of a PLAYLIST (*.m3u) but the name of an MP3 file or stream !
      return AUDIO_FILE_FORMAT_MP3; // not M3U but MP3
   }
  else if( UTL_CheckAndSkipToken( &bp, "application/ogg" ) )  // RFC 3534 - what a misnomer !
   {  // This is the content-type returned from http://67.207.143.181:80/vlf1 :
      return AUDIO_FILE_FORMAT_OGG;
   }
  else if( UTL_CheckAndSkipToken( &bp, "audio/ogg" ) ) // RFC 5334
   {  return AUDIO_FILE_FORMAT_OGG;
   }
  else if( UTL_CheckAndSkipToken( &bp, "audio/vlfrx" ) ) // inoffical...
   {  // Uncompressed stream from VLF-RX-tools *without* T_StreamHeaders
      //  but VT_BLOCKs.  Added here 2020-10-11 but not really supported yet.
      return AUDIO_FILE_FORMAT_STREAM_WITH_VLF_RX_BLOCKS;
   }
  else if( UTL_CheckAndSkipToken( &bp, "text/html" ) ) // RFC 5334
   {  // This is most likely NOT an audio file,
      // but most likely a "404"-like error message from an Icecast server.
      // We still don't know anything until we READ the damned HTML page
      // and try to parse it.
      return iDefaultAudioFileType; // !
   }
  else
   {  return iDefaultAudioFileType;
   }
} // end INET_ContentTypeToAudioFileFormat()

/****************************************************************************/
char * INET_AudioFileFormatToContentType( int iAudioFileType )
{
  static char  sz39Unknown[40];
  switch( iAudioFileType )
   { case AUDIO_FILE_FORMAT_M3U:
        return "audio/x-mpegurl";
     case AUDIO_FILE_FORMAT_MP3:
        return "audio/mpeg";
     case AUDIO_FILE_FORMAT_OGG:
        return "audio/ogg";
     case AUDIO_FILE_FORMAT_STREAM_WITH_HEADERS:
        return "audio/uncompressed"; // this is NOT a recognized "MIME" type; in fact, it's an uncompressed stream identified by T_StreamHeader structs !
     case AUDIO_FILE_FORMAT_STREAM_WITH_VLF_RX_BLOCKS :
        // Uncompressed stream from VLF-RX-tools *without* T_StreamHeaders
        //  but VT_BLOCKs.  Added here 2020-10-11 but not really supported yet.
        return "audio/vlfrx";

     default:
        sprintf( sz39Unknown, "unknown(%d)", (int)iAudioFileType );
        return sz39Unknown;
   }
} // end INET_AudioFileFormatToContentType()


/****************************************************************************/
int INET_SplitURL(
          char * pszFilenameOrURL,   // [in] filename or URL (will be DESTROYED / "chopped into pieces" !)
          T_INET_UrlParts *parts )   // [out] pointers to the 'parts'
  // Breaks up a filename (also for local files), or a URL, into pieces .
  // The ORIGINAL STRING (pszFilenameOrURL) WILL BE DESTROYED !
  // Loosely based on 'HTParse.c' from www.w3.org/Library/src/HTParse.c .
  //
  // [in] pszFilenameOrURL : usually a URL, with the following format :
  //               scheme://domain:port/path?query_string#fragment_id
  //      Example: http://67.207.143.181/vlf6.m3u
  // [out] T_INET_UrlParts *parts : Simple zero-terminated C-strings, example:
  //         parts->access   =
  //         parts->host     = hostname like "qsl.net" or IP string like "192.168.0.123", WITHOUT PORT NUMBER
  //         parts->absolute =
  //         parts->relative =
  //         parts->fragment =
  //         parts->iProtocol: may contain 0=unknown, or INET_PROTOCOL_HTTP, etc .
  //         parts->iPort    : port NUMBER (0=not explicitly set in INET_SplitURL)
  // Return value: ?
{
  int  iResult = 0;
  char * p;
  char * after_access = pszFilenameOrURL;
  memset(parts, 0, sizeof(T_INET_UrlParts) );

  /* Look for fragment identifier */
  if ((p = strchr(pszFilenameOrURL, '#')) != NULL)
   {
     *p++ = '\0';
     parts->pszFragment = p;
   }


  if ((p = strchr(pszFilenameOrURL, ' ')) != NULL)
   { *p++ = '\0';
   }

#ifdef __BORLANDC__
  (void)p;
#endif

  for(p=pszFilenameOrURL; *p; p++)
   {

     // Look for any whitespace. This is very bad for pipelining
     // as it makes the request invalid
     if (isspace((int) *p))
      {
        char *orig=p, *dest=p+1;
        while ((*orig++ = *dest++) != 0);  // scroll up by one character, delete the space character
        p = p-1;
      }
     if (*p=='/' || *p=='#' || *p=='?')
      { break;
      }
     if (*p==':')
      {
        *p = 0;
        parts->pszAccess = after_access; /* Scheme has been specified */

        // The combination of gcc, the "-O" flag and the HP platform is
        // unhealthy. The following three lines is a quick & dirty fix, but is
        // not recommended. Rather, turn off "-O".
        /*              after_access = p;*/
        /*              while (*after_access == 0)*/
        /*                  after_access++;*/

        after_access = p+1;

        if (0==stricmp("URL", parts->pszAccess))
         {
           parts->pszAccess = NULL;  /* Ignore IETF's URL: pre-prefix */
         }
        else
         { break;
         }
      }
   }

  // Added by DL4YHF: If there's this 'access'-thingy, examine it further:
  parts->iProtocol = INET_PROTOCOL_UNKNOWN; // may contain 0=unknown, or INET_PROTOCOL_HTTP, etc .
  if( parts->pszAccess != NULL )
   { if( stricmp(parts->pszAccess, "http" ) == 0 )
      { parts->iProtocol = INET_PROTOCOL_HTTP;
      }
     else if( stricmp(parts->pszAccess, "https" ) == 0 )
      { parts->iProtocol = INET_PROTOCOL_HTTPS;
      }
     else if( stricmp(parts->pszAccess, "file" ) == 0 )
      { parts->iProtocol = INET_PROTOCOL_FILE;  // it's a LOCAL FILE !
      }
     else if( stricmp(parts->pszAccess, "ftp" ) == 0 )
      { parts->iProtocol = INET_PROTOCOL_FTP;
      }
     else if( stricmp(parts->pszAccess, "mailto" ) == 0 )
      { parts->iProtocol = INET_PROTOCOL_FTP;
      }
     else if( stricmp(parts->pszAccess, "rawtcp" ) == 0 )
      { parts->iProtocol = INET_PROTOCOL_RAW_TCP;
      }

   }

  p = after_access;
  if (*p=='/')
   {
          if (p[1]=='/')
      {
        parts->pszHost = p+2;   // host has been specified
        *p=0;                   // Terminate access
        p=strchr(parts->pszHost,'/');   // look for end of host name if any
        if(p)
         {
           *p=0;                                // Terminate host
           parts->pszAbsolute = p+1;   // Root has been found
         }
        // Added by WoBu: If there's a port number after the host name,
        //                separate it, and convert it into integer :
        p=strchr(parts->pszHost,':'); // look for separator between host and port
        if(p)
         { *p=0;             // Terminate host again (now before the port)
           parts->iPort = atoi(p+1);
         }

      }
     else
      {
        parts->pszAbsolute = p+1;               // Root found but no host
      }
   }
  else
   {
     parts->pszRelative = (*after_access) ? after_access : NULL; /* NULL for "" */
   }
  return iResult;
} // end INET_SplitURL()



/****************************************************************************/
BOOL INET_IsInternetResource( char * pszFilenameOrURL )
   // Returns TRUE when pszFilenameOrURL looks like an 'internet' resource;
   //      or FALSE when it's most likely the path+name of a 'local' file .
   // For Spectrum Lab, this function must also recognize stuff beginning with
   //     "raw://", "udp://", "tcp://" as 'internet resources' (like "http://")
   //     even though, strictly/technically, they are NOT ("raw://" isn't
   //     even a known internet protocol, but a dummy for a 'raw' audio stream
   //     on a certain (usually numeric) IP address) .
{
  char *cpFirstColon = strchr(pszFilenameOrURL,':');

  if(  (cpFirstColon != NULL) && (cpFirstColon > (pszFilenameOrURL+2) )
    && (cpFirstColon[1]=='/') && (cpFirstColon[2]=='/')  )
   { // something like http:// ... doesn't look like a file ...
     return TRUE;   // there's colon, but not something like "C:\blabla" ... guess it's an internet stream
   }
  else
   { return FALSE;  // not an (internet-)STREAM but a (local) FILE
   }
} // end INET_IsInternetResource()


/****************************************************************************/
int INET_ReplaceCharInPath( char * pszFilenameOrURL, int iMaxLen, char cOriginal, char cReplacement )
   // Used, for example, to replace '/' in a URL (-path) by '\' for a windoze path,
   //       or vice versa. The old "Unix/Posix versus DOS/Windows" dilemma.
   // Not aware of multi-byte characters !
{
  int nCharsReplaced = 0;
  char *cpEndstop = pszFilenameOrURL + iMaxLen;
  while( (*pszFilenameOrURL != '\0') && (pszFilenameOrURL<cpEndstop) )
   { if(  *pszFilenameOrURL == cOriginal )
      {   *pszFilenameOrURL =  cReplacement;
          ++nCharsReplaced;
      }
     ++pszFilenameOrURL;
   }
  return nCharsReplaced;
} // end INET_ReplaceCharInPath()


/****************************************************************************/
CPROT int INET_InitClient(
     T_INET_Client *pInetClient,    // [out] client data struct
     T_InetCallback pPICallback,    // [in, optional] progress indicator callback
     void *pvUserCallbackData)      // [in, optional] user callback data
{
  if( pInetClient==NULL )
   { return ERROR_ILLEGAL_CMD_PARAMETER;
   }
  memset( pInetClient, 0, sizeof( T_INET_Client ) );
  pInetClient->sock = INVALID_SOCKET;         // note: this is NOT ZERO (thus the need for 'magic')
  pInetClient->pPICallback = pPICallback;
  pInetClient->pvUserCallbackData = pvUserCallbackData;
  pInetClient->iState = INET_CLIENT_STATE_PASSIVE;
  //pInetClient->nMillisecondsWaited = pInetClient->nMillisecondsWaited_peak = 0;
  pInetClient->dwMagic17724538 = 0x17724538;  // declare this struct valid
  pInetClient->dwMagic98696044 = 0x98696044;

  return NO_ERRORS;
} // end INET_InitClient()


/****************************************************************************/
int INET_InvokeCallback(
     T_INET_Client *pInetClient, // [in,out] web client data
     int  iEventType,            // [in] INET_EVENT_TYPE_INFO / ..ERROR ?
     int  iProgressPercent,      // [in] progress in percent, or an 'event code'
     char * pszFormatString,     // [in] format string like "Trying to connect host %s"
     ...)                        // [in] 'value' to be shown, must match the format string
{
  va_list arglist;
  char sz255[256];
  int result=0;

  // Print to string and append to edit control
  va_start(arglist, pszFormatString);
  vsnprintf(sz255, 255, pszFormatString, arglist);
  if( pInetClient->pPICallback != NULL )
   { result = pInetClient->pPICallback( pInetClient->pvUserCallbackData, pInetClient,
                                        iEventType, sz255, iProgressPercent );
   }
  va_end(arglist);  // important to avoid stack corruption - must ALWAYS be called !

  return result;
} // end INET_InvokeCallback()

/****************************************************************************/
static void INET_Client_ProcessHTTPResponse(
              T_INET_Client *pInetClient)    // [in,out] web client data
  // Called after reception of data during state INET_CLIENT_STATE_WAIT_HTTP_RESP .
  // Checks pInetClient->c1kRxBuffer[0...pInetClient->nBytesInRxBuffer-1]
  //        for a valid HTTP response (typically, a HTTP GET response);
  //        and if the response is complete, switches from state
  //        from state INET_CLIENT_STATE_WAIT_HTTP_RESP
  //        into state INET_CLIENT_STATE_WAIT_FILE_DATA .
  // Bytes which don't belong to the HTTP GET response but to the 'file data'
  //        remain in pInetClient->c1kRxBuffer[0...pInetClient->nBytesInRxBuffer-1] .
{
  // Details in http://en.wikipedia.org/wiki/Hypertext_Transfer_Protocol#Server_response :
  // > A server response is followed by a blank line and text of the requested page.
  // >  ...
  // > Most of the header lines are optional. When Content-Length is missing
  // > the length is determined in other ways.
  // Here: Don't care for all those 'header lines', only care for the
  //  'blank line' which separates the response-stuff (with all those chatty
  //  'header lines') from the real "file data", which is what the caller
  //  wants to retrieve .
  //
  // Note: Don't assume anything. Especially not what a 'blank line' looks like.
  //   A blank line may be "\n\n" or "\r\n\r\n" (same stupidity as in text files).
  //
  // Note: We're playing with the internet here, and the internet is often
  //       a crappy, short-living place. No big surprise to receive a response
  //       like the following (from Deutschlandfunk's Ogg Vorbis stream,
  //       http://www.dradio.de/streaming/dlf_mq_ogg.m3u ) :
  //           pInetClient->nBytesInRxBuffer = 242
  //           pInetClient->c1kRxBuffer =
  //    "HTTP/1.0 301 Moved Permanently\r\n"
  //    "Server: squid/2.5.STABLE14\r\n"
  //    "Date: Sun, 11 Dec 2011 21:36:57 GMT\r\n"
  //    "Content-Length: 0\r\n"
  //    "Location: http://www.dradio.de\r\n"
  //    "Set-Cookie: NSC-esbeip_esbeip_fffffffffaseblahdideldumm;"
  //    "path=/;httponly\r\n\r\n" .
  //  Trying the same with Firefox (which sends a much longer GET request)
  //  resulted in an M3U file with the following content (which is the "real" URL):
  //    http://dradio.ic.llnwd.net/stream/dradio_dlf_medium_a.ogg
  //
  int  iIndexOfFirstNonHeaderByte = 0; // buffer-index of the last 'header'-byte : NOT FOUND
  int  i, n, nBytesInRxBuffer = pInetClient->nBytesInRxBuffer;
  char c, *cp, *cpRxBuffer = pInetClient->c1kRxBuffer;
  for(i=0; i<nBytesInRxBuffer-1; ++i)
   { if( cpRxBuffer[i]=='\n' )  // Ok, this may be the end of a line, 'UNIX' or 'DOS' text format....
      { // what's the NEXT character, another '\n', or "\r\n" ?
        if( cpRxBuffer[i+1] == '\n' )  // yes, must be a UNIX-like server ..
         { iIndexOfFirstNonHeaderByte = i+2;   // buffer-index of the last 'header'-byte
           break;
         }
        else
        if((cpRxBuffer[i+1] == '\r' )   // DOS- or WINDOWS text format (with CR+NL as end-of-line markers)
         &&(cpRxBuffer[i+2] == '\n' ) )
         { iIndexOfFirstNonHeaderByte = i+3;   // buffer-index of the last 'header'-byte
           break;
           // Note: Even a server which was definitely running UNIX (or LINUX)
           //       used CR+NL as end-of-line marker in the HTTP responses ;-)
           // The complete HTTP transfer was read in an entire block,
           // including both the HTTP 'GET' response *AND* the file data :
           // cpRxBuffer: "HTTP/1.0 200 OK\r\nContent-Type: audio/x-mpegurl\r\n\r\nhttp://67.207.143.181:80/vlf6\r\n"
         }
      }
   }
  if( iIndexOfFirstNonHeaderByte > 0 ) // found the 'blank line' which marks the end of the response-header-lines !
   { // Since the rx-buffer is 'ours', we can truncate the 'response text' with a zero-byte here,
     // and pass it to the application (just in case the application wants to parse it).
     // In other words, an opportunity for the application to parse the received HTTP GET RESPONSE.
     cpRxBuffer[iIndexOfFirstNonHeaderByte-1] = '\0';
     cp = cpRxBuffer;    // parse the 'header lines' in the HTTP GET response...
     while(*cp != '\0')
      { // Only check for the 'initial response line' and a few 'header lines'
        // (in the HTTP "GET" response), look for a few special tokens .
        // > The header lines are in the usual text header format,
        // > which is: one line per header,
        // >   of the form   "Header-Name: value", ending with CRLF.
        if( UTL_CheckAndSkipToken( (BYTE**)&cp, "HTTP/" ) )
         { // This must be the "Initial Response Line" .
           // Followed by the HTTP protocol version ("1.0" or "1.1", or whatever)
           // and status code as a machine-readable 3-digit integer .
           pInetClient->iHttpMajorVersion = UTL_ParseInteger( (BYTE**)&cp, 1/*digits*/ ); // almost certain the result is ONE
           if( *cp=='.' ) ++cp;
           pInetClient->iHttpMinorVersion = UTL_ParseInteger( (BYTE**)&cp, 1/*digits*/ ); // this result may be ZERO (HTTP 1.0) or ONE (HTTP 1.1)
           pInetClient->iHttpStatusCode = UTL_ParseInteger( (BYTE**)&cp, 3/*digits*/ );   // this result will hopefully be TWOHUNDRED (don't ask why)
         }
        else
        if( UTL_CheckAndSkipToken( (BYTE**)&cp, "Content-Type:" ) )
         { // This thingy defines, more or less, the type of the 'netto data'
           // which will follow immediately after the 'header lines'.
           UTL_SkipSpaces( (BYTE**)&cp );
           UTL_CopyStringUntilEOL( (BYTE*)cp, pInetClient->sz80ContentType, 80 );
         }
        // Increment 'cp' until the end of this line, and skip CR+NL (or only NL?)
        while( (c=*cp++) != '\0')
         { if(c=='\n')  // not 100 % foolproof but reasonably simple
            { break;
            }
         }
      } // end while(*cp != '\0') ...

     if( pInetClient->pPICallback != NULL )
      { pInetClient->pPICallback( pInetClient->pvUserCallbackData, pInetClient,
          INET_EVENT_TYPE_PARSE_RESPONSE, cpRxBuffer, iIndexOfFirstNonHeaderByte );
      }


     // After this, there MAY be bytes remaining in the RX-buffer which belong
     // to the 'netto' file data. We don't want to lose them, thus scroll up
     // the buffer contents, so the first 'file data' byte will be at index ZERO.
     n = pInetClient->nBytesInRxBuffer - iIndexOfFirstNonHeaderByte;
     if( n>0 )  // some "non-header-bytes" will remain in the RX-buffer !
      { for(i=0; i<n; ++i)
         { pInetClient->c1kRxBuffer[i] = pInetClient->c1kRxBuffer[i+iIndexOfFirstNonHeaderByte];
         }
        pInetClient->nBytesInRxBuffer = n;
      }
     else
      { pInetClient->nBytesInRxBuffer = 0;  // the RX-buffer is now "completely empty"
      }

     if( pInetClient->iState == INET_CLIENT_STATE_WAIT_HTTP_RESP )
      {  // had been waiting for reception of the (HTTP-?) response,
         // but after this, we're waiting for the 'file data' :
         pInetClient->iState = INET_CLIENT_STATE_WAIT_FILE_DATA;
      }
   } // end if( iIndexOfFirstNonHeaderByte > 0 )
  else  // Reception of the HTTP GET RESPONSE is obviously not complete yet :
   {
   }
} // end INET_Client_ProcessHTTPResponse()


#if( YHF_INET_USE_SELECT_TO_WAIT )
//---------------------------------------------------------------------------
int INET_WaitForRecv( T_INET_Client *pInetClient, int iTimeout_ms )
  // Call this function after a previous call of recv() is still PENDING,
  //      i.e. after recv() returned SOCKET_ERROR (-1)
  //           but WSAGetLastError() returned WSAEWOULDBLOCK (10035)
  //      which basically means "revc() has not received the number of bytes
  //         that YOU, dear reader, asked it to read". Or something like that.
  //    Thanks to the designers of the Windows Socket Services for making this
  //    so incredibly obfuscated, and for NOT passing a simple timeout value
  //    (in milliseconds) to the recv() function on a BLOCKING socket
  //    (instead, on a blocking socket, recv() would wait for AGES which is
  //     not suited for this application.)
  // Return value:
  //     1 : "ok, successfully waited, Winsock's select() promised there are
  //                data waiting for us now."
  //     0 : "nothing received on the socket, don't waste time, don't call recv()"
  //   < 0 : "something went very wrong on this socket, and I CLOSED IT".
  //          Invert the NEGATIVE return value to get the INET-error code
  //          (not the Winsick-error-code) .
{
  struct timeval timeout = {0};
  int result, iWSAErrorCode, iOptLen;
  T_TIM_Stopwatch myStopwatch;
  int  nMillisecondsWaited;

  timeout.tv_sec  =  iTimeout_ms / 1000;  // Pfui Deibel.. warum einfach, wenn's auch kompliziert geht !
  timeout.tv_usec = (iTimeout_ms % 1000) * 1000;  // kann diese Drecks-Funktion berhaupt 50 Millisekunden warten ?
  // Testen wir doch einfach, wie lange die Drecks-Funktion (select) wirklich wartet,
  // wenn KEINE Daten vom Socket eintrudeln, und auch keine 'exceptional Errors' auftreten !
  TIM_StartStopwatch( &myStopwatch ); // .. to reliably limit the time spent here


  // Forget old flags BEFORE EVERY CALL OF SELECT, at least for the 'set'
  // of sockets on which we want to wait. Seen somewhere on the net:
  // > It's important that you re-enable the file descriptors that were cleared
  // > prior to starting another select, otherwise, you will no longer be
  // > waiting on those file descriptors.
  FD_ZERO(&pInetClient->fdsRead);
  FD_SET( pInetClient->sock, &pInetClient->fdsRead);
     // > FD_SET places sockets into a "set" for various purposes, such as
     // > testing a given socket for readability using the readfds parameter
     // > of the select function.
     // no-no: FD_ZERO(&pInetClient->fdsWrite);
     // no-no: FD_SET( pInetClient->sock, &pInetClient->fdsWrite);
  FD_ZERO(&pInetClient->fdsExceptions);
  FD_SET( pInetClient->sock, &pInetClient->fdsExceptions);
  result = select(0, &pInetClient->fdsRead, NULL, &pInetClient->fdsExceptions, &timeout);
     //           |   |                     |      |
     //           |   | ,-------------------'      |
     //           |   | | ,------------------------'
     //           |   | | |
     //           |   | | An optional pointer to a set of sockets to be checked for errors.
     //           |   | An optional pointer to a set of sockets to be checked for writability.
     //           |   An optional pointer to a set of sockets to be checked for readability.
     //           |
     //           |_________ ignored ("nfds" is a UNIX thing)
     // > The select function returns the total number of socket handles
     // > that are ready and contained in the fd_set structures,
     // > zero if the time limit expired, or SOCKET_ERROR
     // > if an error occurred. If the return value is SOCKET_ERROR,
     // > WSAGetLastError can be used to retrieve a specific error code.
     // > A set of macros is provided for manipulating an fd_set structure.
     // > These macros are compatible with those used in the Berkeley software,
     // > but the underlying representation is completely different.
     // >
     // > The parameter readfds identifies the sockets that are to be checked
     // > for readability. If the socket is currently in the listen state,
     // > it will be marked as readable if an incoming connection request
     // > has been received such that an accept is guaranteed to complete
     // > without blocking. For other sockets, readability means that queued
     // > data is available for reading such that a call to recv, WSARecv,
     // > WSARecvFrom, or recvfrom is guaranteed not to block.
     //
  nMillisecondsWaited = TIM_ReadStopwatch_msec( &myStopwatch );
  if (result < 0)
   { // select() failed.  "This is a permanent error" ... close the socket, give up. Drecksding.
     INET_CloseClient( pInetClient, "after 'select()' failed"/*pszWhy*/ );
     pInetClient->fConnectionLost = TRUE;  // flag for the application
     return -INET_ERROR_SOCKET_NOT_OPEN;  // ERROR: something went direly wrong with the socket, so we CLOSED it
   }
  if (FD_ISSET(pInetClient->sock, &pInetClient->fdsExceptions))
   { // select() told us that some kind of 'exceptional error' has occurred...
     // > The parameter exceptfds identifies the sockets that are to be checked
     // > for the presence of OOB data or any exceptional error conditions.
     iWSAErrorCode = 0;
     iOptLen = sizeof(iWSAErrorCode); // holy crap.. must pass this in BY REFERENCE
     getsockopt( pInetClient->sock, SOL_SOCKET, SO_ERROR, (char*)&iWSAErrorCode, &iOptLen );
     // ( IN SOCKET s, IN int level, IN int optname, OUT char FAR * optval, IN OUT int FAR *optlen)
     INET_CloseClient( pInetClient, "after an 'Exceptional' socket error"/*pszWhy*/ );
     pInetClient->fConnectionLost = TRUE;  // flag for the application
     return -INET_ERROR_SOCKET_NOT_OPEN;  // ERROR: something else went incredibly wrong with the socket, so we CLOSED it
   }
  if (FD_ISSET(pInetClient->sock, &pInetClient->fdsRead))
   { // Bingo, select() informed us about RECEIVED DATA !
     return TRUE;
   }
  else
   { // Wie lange hat das Drecksding (select) WIRKLICH gewartet ?
     // Unter Windows ticken ja alle Uhren sehr ungenau, darum ...
     if( nMillisecondsWaited < (iTimeout_ms-20) )
      { if( ! pInetClient->fProblemsWithTimeSpentInSelect )
         { pInetClient->fProblemsWithTimeSpentInSelect = TRUE; // avoid flooding the log with these...
           INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress or event*/,
              "select() should have waited %d ms, but only waited %d ms.",
              (int)iTimeout_ms, (int)nMillisecondsWaited );
         }
      }
     return FALSE;
   }
} // end INET_WaitForRecv()
#endif // YHF_INET_USE_SELECT_TO_WAIT ?


//---------------------------------------------------------------------------
int INET_OpenAndReadFromServer(
              T_INET_Client *pInetClient,    // [in,out] web client data
              char * pszFilenameOrURL,       // [in] filename or URL
              int  iTimeout_ms,              // [in] max. timeout in milliseconds
              int  iOptions,                 // [in] bitwise combination of the following:
              // INET_RD_OPTION_NORMAL                : return after reception of and number of bytes
              // INET_RD_OPTION_RETURN_AFTER_RESPONSE : return after initial response; no need to wait for the CONTENTS
              // INET_RD_OPTION_WAIT_FOR_ALL_DATA     : return only when all <iMaxLength> bytes received, or timeout
              BYTE *pbDest, int iMaxLength)  // [in,out] destination buffer for the first part of the NETTO DATA
                                             //   (aka "content", not the HTTP response headers)
  // Used, for example, to examine the contents of an M3U file
  //       on a remote server ("stream") .
  // Returns the number of bytes actually placed in pbDest,
  //       or a NEGATIVE ERROR CODE ("error code MADE NEGATIVE") .
  // Note: The error codes in ErrorCodes.h are all POSITIVE !
  //       To tell them from a 'successful result' (NUMBER OF BYTES READ),
  //       we make those errors NEGATIVE here.
  // Note: The TCP/IP connection is NOT automatically closed on return,
  //       because the caller may want to read more bytes from the stream !
{
  struct sockaddr_in sin;
  struct in_addr addr;
  struct hostent *host;
  int result,len,progress_percent;
  int nBytesToSend, nBytesSent;
  int nBytesRcvdNow, nBytesToRead;
  int nBytesInDest = 0;
  int iWSAErrorCode;
  int nMillisecondsWaitedForConnect = 0, nMillisecondsWaitedForData = 0; 
  T_TIM_Stopwatch myStopwatch;
  unsigned long ul;
  char szChoppedURL[INET_MAX_URL_LENGTH+1];
  char sz255Request[256];
  T_INET_UrlParts UrlParts;
  BOOL fRecvdBefore;

# if( YHF_INET_USE_SELECT_TO_WAIT )
  // ex: timeval timeout = {0};   // Borland: "undefined symbol: timeval" .
  // Ok, the last argument of Borland's select() is a "const struct timeval FAR *". Whoo-hoo.
  struct timeval timeout = {0};
  int    iOptLen;  // extravaganza for getsockopt() .
# endif // YHF_INET_USE_SELECT_TO_WAIT ?



#ifdef __BORLANDC__
  (void)nMillisecondsWaitedForConnect;
  (void)nMillisecondsWaitedForData;
#endif


  if( pInetClient==NULL )
   { return -ERROR_ILLEGAL_CMD_PARAMETER;
   }

  if(  (pInetClient->dwMagic17724538 == 0x17724538)
    && (pInetClient->dwMagic98696044 == 0x98696044)
    && (pInetClient->sock != INVALID_SOCKET) )  // oops ?!
   { // both magic numbers valid (pi^0.5, pi^2) -> was open but not closed yet !!
     // bear with the caller; close old connection autoMAGICally
     INET_CloseClient( pInetClient, "before re-opening"/*pszWhy*/ );
     pInetClient->sock = INVALID_SOCKET; // redundant but defensive (don't close twice)
   }

  if( pszFilenameOrURL==NULL )
   { return -1;  // ERROR: something terribly wrong with the input arguments
   }

  len = strlen(pszFilenameOrURL);
  if( len<1 || len>INET_MAX_URL_LENGTH )
   { return -1;  // ERROR: something terribly wrong with the input arguments
   }
  strcpy(szChoppedURL, pszFilenameOrURL );
  result = INET_SplitURL( szChoppedURL, &UrlParts );
#ifdef __BORLANDC__
  (void)result;
#endif

  if( UrlParts.pszHost == NULL )
   { return -INET_ERROR_BAD_URL_SYNTAX;  // ERROR: not a valid "host"-name
   }
  pInetClient->iProtocol = UrlParts.iProtocol;
  pInetClient->iState = INET_CLIENT_STATE_PASSIVE;
  pInetClient->nBytesInRxBuffer = 0;
  pInetClient->ulTotalBytesReceived = 0; // for diagnostic purposes
  pInetClient->ulTotalBytesSent     = 0;

  if( INET_IsInternetResource( pszFilenameOrURL ) )
   { // it's an internet resource, not a FILE ON THE LOCAL DISK:
     pInetClient->is_internet_resource = TRUE;
     pInetClient->sock = socket(AF_INET, SOCK_STREAM, 0);  /* init socket descriptor */
     if( pInetClient->sock==INVALID_SOCKET )
      { INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress*/,
           "Failed to create socket" );
        return -INET_ERROR_SOCKET_NOT_OPEN;  // ERROR: something went wrong with the socket
        // Note that the error codes in ErrorCodes.h are all POSITIVE !
        // To tell them from a 'successful result' (NUMBER OF BYTES READ),
        // we make those errors NEGATIVE here.  Any better solution ?
      }
#   if( YHF_INET_USE_SELECT_TO_WAIT )  // even stranger things to "select()" [="wait for something"]
     FD_ZERO(&pInetClient->fdsRead);
     FD_SET( pInetClient->sock, &pInetClient->fdsRead);
     // > FD_SET places sockets into a "set" for various purposes, such as
     // > testing a given socket for readability using the readfds parameter
     // > of the select function.
     FD_ZERO(&pInetClient->fdsWrite);
     FD_SET( pInetClient->sock, &pInetClient->fdsWrite);
     FD_ZERO(&pInetClient->fdsExceptions);
     FD_SET( pInetClient->sock, &pInetClient->fdsExceptions);
#   endif // YHF_INET_USE_SELECT_TO_WAIT ?

     // Note that some say gethostbyname does not resolve IP address strings passed to it !
     // Such a request is treated exactly as if an unknown host name were passed.
     // An application with an IP address string to resolve should use inet_addr
     // to convert the string to an IP address, then gethostbyaddr to obtain the hostent structure.
     // Anyway, with UrlParts.pszHost == "67.207.143.181", gethostbyname() returned NON-NULL !
     //   Maybe it's a stupid IP STRING (like "192.168.0.222"),
     //   so we don't need to look up a 'host by name' at all:
     addr.s_addr = inet_addr(UrlParts.pszHost); // maybe it's an "IP-String"..
     host = gethostbyaddr((char*)&addr, sizeof(addr), AF_INET);  // fails very often - WHY ?
     if( host==NULL )
      { // Obviously not a simple "dotted IP string", so:
        host = gethostbyname(UrlParts.pszHost);
        // > If no error occurs, gethostbyname returns a pointer
        // > to the hostent structure described above. Otherwise, it returns a NULL pointer.
      }
     if( host==NULL ) // neither a stupid IP STRING ("192.168.0.222"), nor a known host name !
      { closesocket( pInetClient->sock );  pInetClient->sock=INVALID_SOCKET;
        INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress*/,
           "gethostbyname(%s) failed", UrlParts.pszHost );
        return -INET_ERROR_SOCKET_NOT_OPEN;  // ERROR: something else went wrong with the socket
      }

     // Prepare the sockaddr_in structure which we need to connect to the remote host
     memcpy(&sin.sin_addr.s_addr, host->h_addr, host->h_length);
        // Note: There isn't a h_addr member in the 'hostent' structure.
        //  But the compiler didn't complain. Is it a stupid MACRO ?
        //  Yes it is. #define h_addr  h_addr_list[0]. No surprise that the debugger
        //  cannot evaluate it.  The h_addr_list[]-thingy is just another monster,
        //  it's actually a char FAR * FAR * ;  (forget about the stoneage-"FAR")
        //  This is just another annoyance of the obfuscated socket library.
        //  To inspect this obfuscated mess with a debugger, WATCH this:
        //  (int)(unsigned char)host->h_addr_list[0][0], ..[0][1], ..[0][2], ..[0][3].
        //  The result should be the IP address .
     sin.sin_family = AF_INET;
     if( UrlParts.iPort<=0 )
      {
        switch( pInetClient->iProtocol )
         { case INET_PROTOCOL_HTTP:
           default:
              UrlParts.iPort = 80;  // default for HTTP
              break;
         } // end switch( pInetClient->iProtocol )
      }
     sin.sin_port = htons(UrlParts.iPort);

     // Connect to the socket service described by the sockaddr_in struct :
     INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_INFO|INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress*/,
         "Trying to connect host %s", UrlParts.pszHost );
     // The default 'blocking' connect() wasted up to TEN SECONDS,
     //  sometimes even more, until returning with an error .
     //  As so often with the crippled 'Winsock', there's no option
     //  (with 'setsockopt()') to set the CONNECT-timeout.
     //  You can only set the RECEIVE- and TRANSMIT-timeout. What a dog.

#  if( ! YHF_INET_USE_SELECT_TO_WAIT )
     // Old stuff (always in use before 2020-10-17) :
     //     Call connect() BEFORE making the socket non-blocking .
     // Note: It seems important to 'connect' BEFORE making the socket non-blocking,
     //       otherwise connect() will usually fail, and there's no way
     //       to instruct 'connect' to NOT return until the connection
     //       is established. To keep it simple, do NOT use a loop here:
     if (connect(pInetClient->sock, (struct sockaddr *)&sin, sizeof(sin)) < 0)
      { iWSAErrorCode = WSAGetLastError();
        closesocket( pInetClient->sock );   pInetClient->sock=INVALID_SOCKET;
        INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress*/,
             "Failed to connect %s : %s", UrlParts.pszHost,
             INET_WinsockErrorCodeToString(iWSAErrorCode) );
        return -INET_ERROR_COULDNT_CONNECT;
      }
     INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_INFO|INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress*/,
         "Connected to %s", UrlParts.pszHost );
#  endif // ! YHF_INET_USE_SELECT_TO_WAIT ?

     // Ex: AFTER successfully connecting the remote "host" (server),
     //      try to make this socket "non-blocking".
     //      This is not very clear in the Win32 programmer's reference !
#if(1)
     // FIONBIO: 'Enable or disable nonblocking mode on socket' .
     ul = 1;  // nonzero for nonblocking mode
     result = ioctlsocket( pInetClient->sock, FIONBIO, &ul );
     if(result == SOCKET_ERROR)
      { iWSAErrorCode = WSAGetLastError();
        closesocket( pInetClient->sock );
        pInetClient->sock=INVALID_SOCKET;
        INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR, 0/*progress or event*/,
           "Failed to make socket non-blocking : %s",
           INET_WinsockErrorCodeToString(iWSAErrorCode) );
        return -INET_ERROR_SOCKET_NOT_OPEN;  // ERROR: something else went wrong with the socket
      }
#endif

#  if( YHF_INET_USE_SELECT_TO_WAIT )
     // New since 2020-10-17 :
     //     connect() AFTER making the socket non-blocking, using
     //     WSAGetLastError(), select() [with a timeout passed as argument],
     //     and the usual bunch of non-intuitive Winsock API functions ...
     TIM_StartStopwatch( &myStopwatch ); // RELIABLY(!) measure the time waiting for connect()..
     if( connect(pInetClient->sock, (struct sockaddr *)&sin, sizeof(sin)) == SOCKET_ERROR)
      { // Something went wrong, and we need yet another API function to find out what:
        if( (iWSAErrorCode=WSAGetLastError()) != WSAEWOULDBLOCK )
         { // It's a REAL error during the NON-BLOCKING call of connect(). Shit happens !
           closesocket( pInetClient->sock );
           pInetClient->sock=INVALID_SOCKET;
           INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress or event*/,
              "Non-blocking connect() failed immediately : %s",
              INET_WinsockErrorCodeToString(iWSAErrorCode) );
           return -INET_ERROR_SOCKET_NOT_OPEN;  // ERROR: something else went wrong with the socket
         }
        // Arrived here: connect() promised to try, but the connection is still pending.
        // Time to WAIT for completion, using yet another bunch of obfuscated API functions.
        // Most people suggest to wait for BEING ABLE TO WRITE (via select) here after connect(),
        // because the guy at the other side may not send anything, so there's
        // no point in passing &pInetClient->fdsRead to select() here:
        timeout.tv_sec  = (iTimeout_ms + 999) / 1000;
        timeout.tv_usec = 0;
        result = select(0, NULL, &pInetClient->fdsWrite, &pInetClient->fdsExceptions, &timeout);
        if (result <= 0)
         { // select() failed or connection timed out
           closesocket( pInetClient->sock );
           pInetClient->sock=INVALID_SOCKET;
           INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress or event*/,
              "Non-blocking connect() timed out" );
           return -INET_ERROR_SOCKET_NOT_OPEN;  // ERROR: something else went wrong with the socket
         }
        if (FD_ISSET(pInetClient->sock, &pInetClient->fdsExceptions))
         { // connection failed
           iWSAErrorCode = 0;
           iOptLen = sizeof(iWSAErrorCode); // holy crap.. must pass this in BY REFERENCE
           getsockopt( pInetClient->sock, SOL_SOCKET, SO_ERROR, (char*)&iWSAErrorCode, &iOptLen );
           // ( IN SOCKET s, IN int level, IN int optname, OUT char FAR * optval, IN OUT int FAR *optlen)

           closesocket( pInetClient->sock );
           pInetClient->sock = INVALID_SOCKET;
           INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress or event*/,
              "Failed to connect %s : %s",
              UrlParts.pszHost,
              INET_WinsockErrorCodeToString(iWSAErrorCode) );
           return -INET_ERROR_SOCKET_NOT_OPEN;  // ERROR: something else went wrong with the socket
         }
        if (! FD_ISSET(pInetClient->sock, &pInetClient->fdsWrite))
         { // select() returned but our socket isn't ready to WRITE ? ?
           closesocket( pInetClient->sock );
           pInetClient->sock=INVALID_SOCKET;
           INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress or event*/,
              "Freshly connected socket not ready to 'write' - closing." );
           return -INET_ERROR_SOCKET_NOT_OPEN;  // ERROR: something else went wrong with the socket
         } // end if < socket not ready to WRITE >
       }
      else // non-blocking connect() returned IMMEDIATELY, WITH success ?
       {   // Guess this never happens. Anyway, it's windows, don't assume anything.
       }
      nMillisecondsWaitedForConnect = TIM_ReadStopwatch_msec( &myStopwatch );

      // Connection successful.   REALLY ? Oh boy, only a few dozen lines of code to check for success !
      INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_INFO|INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress*/,
         "Connected to %s within %d ms .", UrlParts.pszHost, (int)nMillisecondsWaitedForConnect );
      // Real-world example, source = tcp://129.206.29.82:49335 :
      //    Connected to 129.206.29.82 within 284 ms (1st call, for detection of the stream format);
      //    Connected to 129.206.29.82 within 284 ms (2nd call, from VorbisFileIO.cpp).
      //
#  endif // YHF_INET_USE_SELECT_TO_WAIT ?


     pInetClient->fConnectionLost = FALSE;  // clear this 'old' flag for the application

     // If the protocol is HTTP (which will often be the case when starting
     //  to read an audio stream hidden in an M3U "playlist") ,
     //  the client (i.e. THIS END) must inform the remote server
     //  what to do. In HTTP parlance: Send a GET command, plus some red tape
     //  (barebone web programming always comes along with an AWFUL lot of red tape..)
     // From a simple introduction on HTTP from http://www.jmarshall.com/easy/http/#requestline :
     // > To retrieve the file at the URL
     // >     http://www.somehost.com/path/file.html
     // > first open a socket to the host www.somehost.com, port 80
     // > (use the default port of 80 because none is specified in the URL).
     // > Then, send something like the following through the socket:
     // >     GET /path/file.html HTTP/1.0
     // >     From: someuser@jmarshall.com
     // >     User-Agent: HTTPTool/1.0
     // >     [blank line here]
     // > The server should respond with something like the following,
     // > sent back through the same socket:
     // >     HTTP/1.0 200 OK
     // >     Date: Fri, 31 Dec 1999 23:59:59 GMT
     // >     Content-Type: text/html
     // >     Content-Length: 1354
     // >     <html> ........
     // From Wikipedia :
     // > The request message consists of the following:
     // >
     // >  *  Request line, such as GET /images/logo.png HTTP/1.1,
     // >     which requests a resource called /images/logo.png from server
     // >  *  Headers, such as Accept-Language: en
     // >  *  An empty line.
     // >  *  An optional message body.
     // > The request line and headers must all end with <CR><LF>
     // > (that is, a carriage return followed by a line feed).
     // > The empty line must consist of only <CR><LF> and no other whitespace .
     pInetClient->iState = INET_CLIENT_STATE_SENDING_REQUEST;
     switch( pInetClient->iProtocol )
      { case INET_PROTOCOL_HTTP :
           strcpy( sz255Request, "GET " );
           if( UrlParts.pszAbsolute != NULL )
            { if( UrlParts.pszAbsolute[0] != '/' )
               { strcat( sz255Request, "/" );
               }
              strcat( sz255Request, UrlParts.pszAbsolute );
            }
           else if( UrlParts.pszRelative != NULL )
            { strcat( sz255Request, UrlParts.pszRelative );
            }
           strcat( sz255Request, " HTTP/1.0\r\n\r\n" );   // note the 2nd -EMPTY- line !
           nBytesToSend = strlen(sz255Request);
           nBytesSent = send( pInetClient->sock, sz255Request, nBytesToSend, 0 );
           if( nBytesSent != nBytesToSend )
            { INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress or event*/,
                     "Failed to send GET-request" );
            }
           else
            { pInetClient->iState = INET_CLIENT_STATE_WAIT_HTTP_RESP;
              INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_INFO, 1/*progress or event*/,
                     "%s", sz255Request );
            }
           break; // end case INET_PROTOCOL_HTTP
         case INET_PROTOCOL_RAW_TCP :  // added 2011-12-11 to experiment with timestamped Ogg/Vorbis streams
           // Nothing to send in this case (no REQUEST);
           //  the "file" data (in fact, a stream) should just start to pour in !
           pInetClient->iState = INET_CLIENT_STATE_WAIT_FILE_DATA;
           break; // end case INET_PROTOCOL_RAW_TCP
         default:
           break;
      } // end switch( pInetClient->iProtocol )

     // Immediately READ SOMETHING (HTTP response, "hello from the server" or similar)
     //      after opening the connection ?
     if( (pbDest!=NULL) && (iMaxLength>0) )
      {
        // Try to read the first NNN bytes from the remote host.
        // Note that, at this point, the socket is NON-BLOCKING which means
        // revc() will return IMMEDIATELY if there are no data waiting
        // in the network buffer !
        if( iTimeout_ms < 10 )  // avoid div-by-zero, and allow at least 10 ms
         {  iTimeout_ms = 10;
         }
        TIM_StartStopwatch( &myStopwatch ); // forget the time spent waiting for connect()
        fRecvdBefore = FALSE;
        INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_INFO|INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress*/,
           "Waiting for reception of first data" );

        while ( pInetClient->iState == INET_CLIENT_STATE_WAIT_HTTP_RESP )
         {
           progress_percent = (100 * nMillisecondsWaitedForData) / iTimeout_ms;
           nBytesToRead = sizeof(pInetClient->c1kRxBuffer) - pInetClient->nBytesInRxBuffer;
           nBytesRcvdNow = recv( // why do so many people use 'read' instead of 'recv' ?
                 pInetClient->sock,   // [in] socket; descriptor identifying a connected socket
                 pInetClient->c1kRxBuffer // [out] buffer for the received HTTP response (or similar)
                  + pInetClient->nBytesInRxBuffer,
                 nBytesToRead,        // [in] MAX. number of bytes acceptable in that buffer
                 0 );                 // [in] flags; specifies how the call is made
              // > If no error occurs, recv returns the number of bytes received.
              // > If the connection has been gracefully closed, the return value
              // > is zero. Otherwise, a value of SOCKET_ERROR is returned,
              // > and a specific error code can be retrieved by calling WSAGetLastError.
              // > If no incoming data is available at the socket,
              // > the recv call waits for data to arrive
              // > unless the socket is nonblocking. In this case, a value of
              // > SOCKET_ERROR (-1) is returned with the error code set to WSAEWOULDBLOCK.
              // > The select, WSAAsyncSelect, or WSAEventSelect calls can be used
              // > to determine when more data arrives.
           if( nBytesRcvdNow > 0 )  // hooray, successfully "received something" !
            { pInetClient->nBytesInRxBuffer += nBytesRcvdNow;
              pInetClient->ulTotalBytesReceived += nBytesRcvdNow;
              fRecvdBefore = TRUE;
              INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_INFO|INET_EVENT_FLAG_SHOW_TIMESTAMP, progress_percent,
                   "Receiving response.. %d bytes", (int)pInetClient->nBytesInRxBuffer );
              // Still waiting for the reception of the HTTP RESPONSE,
              //       or already receiving the netto "file data" ?
              INET_Client_ProcessHTTPResponse( pInetClient );
              // Note: INET_Client_ProcessHTTPResponse() may switch the state
              //       from INET_CLIENT_STATE_WAIT_HTTP_RESP
              //         to INET_CLIENT_STATE_WAIT_FILE_DATA,
              //       which would terminate this 'waiting loop' .
              if( pInetClient->iState != INET_CLIENT_STATE_WAIT_HTTP_RESP )
               { // Finished waiting for the GET-response ("headers"),
                 // may already have read the first part of the "netto file data"
                 // because the network buffer won't care for 'empty lines'
                 // after the HTTP GET header(s) !
                 if( iOptions & INET_RD_OPTION_RETURN_AFTER_RESPONSE )
                  { // return after initial response; no need to wait for the CONTENTS
                    // (despite that, bytes which don't belong to the HTTP GET response
                    //   -but to the 'file data' aka 'content'- remain in
                    //  pInetClient->c1kRxBuffer[0...pInetClient->nBytesInRxBuffer-1] )
                    break;
                  }
                 else
                  { nBytesInDest = INET_ReadDataFromServer( pInetClient,
                                    iTimeout_ms, iOptions, pbDest, iMaxLength );
                  }
               } // end if( pInetClient->iState != INET_CLIENT_STATE_WAIT_HTTP_RESP )
            }
           else // nothing received this time, or connection closed, or another ERROR ?
            { if( nBytesRcvdNow==SOCKET_ERROR )  // obviously a REAL ERROR... or what ?!!??!!?!
               { iWSAErrorCode = WSAGetLastError();
                 if( iWSAErrorCode == WSAEWOULDBLOCK ) // winsock's way of saying 'nothing received' ?
                  { // Since the socket is NON-BLOCKING, don't call recv() on it again.
                    // Instead, use select() to WAIT for a few milliseconds ?
                    pInetClient->fRecvPending = TRUE;
                  }
                 else // not just 'nothing received' but a REAL ERROR.  Which ?
                  { INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, progress_percent,
                      "Error from recv: %s",
                      INET_WinsockErrorCodeToString(iWSAErrorCode) );
                    break;
                  }
               }
              else  // nothing received, but no "socket error"
               { // Got here even with a BLOCKING socket(!),
                 // which did NOT block, but returned ZERO !
               }
              if( fRecvdBefore )  // nothing received now, but something received before
               { // -> network buffer is completely empty; "done" !
                 break;
               }
              else
               { // Wait a few milliseconds more, typically for the 'initial response'
                 if( iTimeout_ms>0 )
                  { // Give the remote server ("host") time to reply
                    // ex: Sleep(50/*ms*/);
                    result = INET_WaitForRecv( pInetClient, 10/*iTimeout_ms*/ );
                    // As so often, things may go wrong where you least expect it.
                    // For example, select() may have noticed an 'exceptional error'
                    // with a socket. In that case, INET_WaitForRecv() may have
                    // already CLOSED the socket and returned with a negative-made
                    // error code (e.g. -INET_ERROR_SOCKET_NOT_OPEN)
                    if( result < 0 )  // 'exceptional error' somewhere in INET_WaitForRecv()..
                     { break;
                     }
                  }
                 else // call is made "completely non-blocking" :
                  { break;
                  }
               }
            }
           if( pInetClient->sock==INVALID_SOCKET ) // oops.. someone CLOSED the socket !
            { break;
            }
           nMillisecondsWaitedForData = TIM_ReadStopwatch_msec( &myStopwatch );
           if( nMillisecondsWaitedForData > iTimeout_ms )
            { // been waiting too long (at least longer than the caller can accept) ->
              INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress or event*/,
                  "Timed out (%d ms) waiting for response", (int)nMillisecondsWaitedForData );
              break;
            }
         } // end while
      } // end if < receive something IMMEDIATELY after connect() > ?
     // Note: The connection REMAINS OPEN !
   } // end if < not a file but an internet resource, most likely a STREAM or M3U on remote server >
  else
   { // it's a FILE on the local disk:
     pInetClient->is_internet_resource = FALSE;
     pInetClient->fConnectionLost = FALSE;
   } // end else < file on the local disk>

  if( iMaxLength > 0 )
   { INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_INFO | INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress*/,
       "OpenAndRead(\"%s\") returns %d bytes, wanted: %d, %d ms .",
       (char*)pszFilenameOrURL, (int)nBytesInDest, (int)iMaxLength, (int)nMillisecondsWaitedForData );
   }
  else // caller didn't want to READ anything, so don't say "returned 0 bytes":
   { INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_INFO | INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress*/,
       "OpenAndRead(\"%s\") successful", (char*)pszFilenameOrURL );
   }

  return nBytesInDest;
} // end INET_OpenAndReadFromServer()

/****************************************************************************/
int INET_ReadDataFromServer(  // ... usually a BLOCKING call .
     T_INET_Client *pInetClient,    // [in,out] web client data
     int  iTimeout_ms,              // [in] max. timeout in milliseconds
     int  iOptions,                 // [in] bitwise combination of the following:
     // INET_RD_OPTION_NORMAL                : return after reception of any number of bytes
     // INET_RD_OPTION_WAIT_FOR_ALL_DATA     : return only when all <iMaxLength> bytes received, or timeout
     BYTE *pbDest, int iMaxLength)  // [in,out] destination buffer
  // Used, for example, to examine the contents of an M3U file
  // on a remote server ("stream") .
  // Returns the number of bytes actually placed in pbDest,
  //       or a NEGATIVE ERROR CODE ("error code MADE NEGATIVE") .
  // Note: The error codes in ErrorCodes.h are all POSITIVE !
  //       To tell them from a 'successful result' (NUMBER OF BYTES READ),
  //       we make those errors NEGATIVE here.
  // Note: The TCP/IP connection is NOT automatically closed on return,
  //       because the caller may want to read more bytes from the stream !
{
  int i,n,nBytesRcvdNow;
  int nBytesToRead, nBytesRead = 0, nLoops = 0;
  int nMillisecondsWaited = 0;
  int result, iWSAErrorCode;
  T_TIM_Stopwatch myStopwatch;
  BOOL fRecvdBefore;
  unsigned long ulPrevTotalBytesReceived = pInetClient->ulTotalBytesReceived;

  if( pInetClient->sock==INVALID_SOCKET )
   { return -INET_ERROR_SOCKET_NOT_OPEN;
   }

  if( pInetClient->is_internet_resource )
   { if( pInetClient->iState != INET_CLIENT_STATE_WAIT_HTTP_RESP )
      { // Finished waiting for the GET-response ("headers"),
        // may already have read the first part of the "netto file data"
        // because the network buffer won't care for 'empty lines'
        // after the HTTP GET header(s) !
        // Bytes which don't belong to the HTTP GET response
        // but to the 'file data' remain in
        // pInetClient->c1kRxBuffer[0...pInetClient->nBytesInRxBuffer-1] .
        if( pInetClient->nBytesInRxBuffer > 0 )
         { // before attempting to RECEIVE more (from the socket),
           // deliver those data from the internal RX-buffer :
           nBytesToRead = pInetClient->nBytesInRxBuffer;
           if( nBytesToRead > iMaxLength )
            {  nBytesToRead = iMaxLength;
            }
           if( nBytesToRead > 0 )
            {  memcpy( pbDest, pInetClient->c1kRxBuffer, nBytesToRead );
               pbDest     += nBytesToRead;
               nBytesRead += nBytesToRead;
               iMaxLength -= nBytesToRead;
               if( iMaxLength>0 )
                { *pbDest = 0;  // append a trailing zero, as a service for TEXT STRINGS
                }
               n = pInetClient->nBytesInRxBuffer - nBytesToRead;
               if( n>0 )  // some "non-header-bytes" still remain in the RX-buffer !
                { for(i=0; i<n; ++i)   // (this is unusual but .. who knows)
                   { pInetClient->c1kRxBuffer[i] = pInetClient->c1kRxBuffer[i+nBytesToRead];
                   }
                  pInetClient->nBytesInRxBuffer = n;
                }
               else  // the RX-buffer is now "completely empty"; no need to "move up" !
                { pInetClient->nBytesInRxBuffer = 0;
                }
            } // end if( nBytesToRead > 0 )
         } // end if( pInetClient->nBytesInRxBuffer > 0 )
        else  // no "old" bytes remaining in the internal rx-buffer :
         { nBytesRead = nBytesRead;
         }
      } // end if( pInetClient->iState != INET_CLIENT_STATE_WAIT_HTTP_RESP )
   }

  if( iMaxLength>0 )  // still more bytes 'acceptable' by the caller ?
   { // Try to read the next N bytes from the remote host (file "contents") .
     // Note that, at this point, the socket is NON-BLOCKING which means
     // revc() will return IMMEDIATELY if there are no data waiting
     // in the network buffer !
     TIM_StartStopwatch( &myStopwatch ); // RELIABLY(!) limit the time spent below
     fRecvdBefore = FALSE;
     nMillisecondsWaited = 0;
     while( (nMillisecondsWaited <= iTimeout_ms)  &&  (iMaxLength>0) )
      { ++nLoops;
        nBytesRcvdNow = recv( // <- a NON-BLOCKING call.. returns immediately !
              pInetClient->sock,   // [in] socket; descriptor identifying a connected socket
              (char*)pbDest,       // [out] (remaining) buffer for the incoming data
              iMaxLength,          // [in] (remaining) length of the above buffer
              0 );                 // [in] flags; specifies how the call is made
        // Because our socket is NON-BLOCKING, recv() will usually NOT return
        // anything yet, and return nBytesRcvdNow=-1 ... more on that further below.
        if( nBytesRcvdNow>0 )  // hooray, successfully "received something" !
         { pInetClient->ulTotalBytesReceived += nBytesRcvdNow;
           nBytesRead += nBytesRcvdNow;
           pbDest     += nBytesRcvdNow;
           iMaxLength -= nBytesRcvdNow;
           if( iMaxLength>0 )
            { *pbDest = 0;  // append a trailing zero, as a service for TEXT STRINGS
            }
           fRecvdBefore = TRUE;
         }
        else // nothing received this time; wait for a few more milliseconds ?
         { if( nBytesRcvdNow < 0 ) // a NEGATIVE return from recv() indicates an error.. which ?
            { iWSAErrorCode = WSAGetLastError(); // how does this crap know for WHICH SOCKET we're calling it ?
              if( iWSAErrorCode == WSAEWOULDBLOCK )  // winsock's way of saying 'nothing received' ?
               { // Not really an error. As usual, the documentation is poor.
                 // recv() often returned ZERO when nothing was received.
                 nBytesRcvdNow = 0;
#               ifdef __BORLANDC__
                 (void)nBytesRcvdNow; // "... is assigned a value that is never used" - so what ?
#               endif

               }
              else // not just 'nothing received' but a REAL ERROR.  Which ?
               { INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, iWSAErrorCode,
                      "Error from recv: %s",
                      INET_WinsockErrorCodeToString(iWSAErrorCode) );
                 break;
               }
            }
           if( fRecvdBefore  // nothing received now, but something received before
             && ( (iOptions & INET_RD_OPTION_WAIT_FOR_ALL_DATA) == 0 ) )
            { // -> guess the network buffer is empty now; "done" !
              break;
            }
           else
            { // Wait a few milliseconds more, typically for the 'initial response'.
              // CAUTION: Sleeping here often doesn't help, especially not when
              // called from a high-priority worker thread like TSoundThread::Execute !
              if( iTimeout_ms > 0 )
               {
                 // .. give the remote server ("host") time to reply or send more data:
                 // ex: Sleep(50/*ms*/);
                 result = INET_WaitForRecv( pInetClient, 10/*iTimeout_ms*/ );
                 // As so often, things may go wrong where you least expect it.
                 // For example, select() may have noticed an 'exceptional error'
                 // with a socket. In that case, INET_WaitForRecv() may have
                 // already CLOSED the socket and returned with a negative-made
                 // error code (e.g. -INET_ERROR_SOCKET_NOT_OPEN)
                 if( result < 0 )  // 'exceptional error' somewhere in INET_WaitForRecv()..
                  { break;
                  }
               }
              else // call is made "completely non-blocking" :
               { break;
               }
            } // end else < WAIT (for the arrival of new data, a few ms >
         } // end else < nothing received THIS TIME >
        nMillisecondsWaited = TIM_ReadStopwatch_msec( &myStopwatch ); // don't rely on select() as a "timer" !
      } // end while < more bytes to read from the remote host >
   } // if( iMaxLength>0 )

  // 2016-01 : When trying to receive a few hundred bytes from Jacek's (SQ5BPF)
  //           test server, occasionally only received SIXTEEN bytes, despite
  //           a timeout value of 2500 milliseconds .
  //           Increased to 3000 ms (in TSpectrumLab::OpenInputFileAnalysis) .
  nLoops = nLoops;  // << potential place for a breakpoint
#ifdef __BORLANDC__
  (void)nLoops; // "... is assigned a value that is never used" .. shut up !
#endif


  if( (ulPrevTotalBytesReceived==0) && (pInetClient->ulTotalBytesReceived>0) )
   {
     INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_INFO | INET_EVENT_FLAG_SHOW_TIMESTAMP, 0/*progress*/,
           "Received first %d bytes after %d ms .",
           (int)pInetClient->ulTotalBytesReceived, (int)nMillisecondsWaited );
     // Real-world example, source = tcp://129.206.29.82:49335 (DK7FC) :
     //  > Received first 1024 bytes after 349 ms (1st call, for detection of the stream format);
     //  > Received first 65536 bytes after 316 ms (2nd call, from VorbisFileIO.cpp).
     //
   }

  return nBytesRead;
} // end INET_ReadDataFromServer()


/****************************************************************************/
static int INET_SendImmediately(  // "continues" writing/sending if necessary
     T_INET_Client *pInetClient,    // [in,out] web client data
     int  iTimeout_ms,              // [in] max. timeout in milliseconds (0=non-blocking)
     int  iOptions,                 // [in] combination of INET_WR_OPTION_....
     BYTE *pbSource, int iLength)   // [in] source data, and number of bytes to send
  // Beware of the Nagle algorithm .. only send as large blocks as possible !
  // Return value :
  //      number of bytes actually written,
  //  or ZERO if the network driver's buffer is full (send->WSAEWOULDBLOCK)
  //  or a NEGATIVE error code
  //     compatible with c:\cbproj\SoundUtl\ErrorCodes.cpp::ErrorCodeToString().
{
  int iErrorCodeOrNumBytesWritten = 0;
  int nLoops = 20;   // don't try to send() more than this number of times
  int n;
  // int nBytesToRead, nBytesRead = 0;
  int nMillisecondsWaited = 0;
  int iWSAErrorCode;
  // BOOL fRecvdBefore;

  if( pInetClient->sock==INVALID_SOCKET )
   { return -INET_ERROR_SOCKET_NOT_OPEN;
   }


  if( (pbSource!=NULL) && (iLength>0) )  // is there really something to send ?
   {
     if( pInetClient->is_internet_resource )
      { // It's indeed an 'internet' connection, most likely TCP/IP-based ...
        switch( pInetClient->iProtocol )
         { case INET_PROTOCOL_HTTP    :
           case INET_PROTOCOL_RAW_TCP :
              while( (iLength > 0) && (nLoops>0) )
               { --nLoops;
                 n = send( pInetClient->sock, (char*)pbSource, iLength, 0/*flags*/ );
                 // Note: Use 'TCP Dummy Sink for Audio Streams' to check the
                 //       block side RECEIVED at the other side !
                 //       Blocks of over 8192 bytes seemed to arrive in a
                 //       single call of recv() when tested in the LAN .
                 if( n>0 )  // ok, successfully "sent something" (not necessarily all !)
                  { pInetClient->ulTotalBytesSent += n;
                    iLength -= n;
                    pbSource+= n;
                    iErrorCodeOrNumBytesWritten += n;
                  }
                 else // nothing sent this time; wait a few milliseconds more.. it's a NON-BLOCKING socket !
                  {
                    iWSAErrorCode = WSAGetLastError();
                    // Got here REPEATEDLY with iWSAErrorCode == 10035 == WSAEWOULDBLOCK;
                    // no matter how long we waited. That's not what a socket should do !
                    switch( iWSAErrorCode )
                     { case WSAEWOULDBLOCK : // winsock's way of saying 'TX-buffer it full' ?
                          // Not really an error. Most likely, the network buffer is full .
                          // If the caller wants to, wait a few milliseconds here.
                          // 2012-04-25 : Got here when launching google earth :)
                          if( iTimeout_ms > 0 )
                           {
                             // Give the network stack some time to think & send:
                             Sleep(20/*ms*/);
                             nMillisecondsWaited += 20/*ms*/;
                           }
                          else // call is made "completely non-blocking" :
                           { // This likely happens in Cumiana regularly,
                             // when the socket's own network buffer is exhausted
                             // an SPECTRUM LAB needs to fill up its own buffer.
                             // THIS IS NOT AN ERROR YET !
                             // The caller (INET_Write) will decide.
                             // No-No: iErrorCodeOrNumBytesWritten = -INET_ERROR_CANNOT_SEND;
                             nLoops = 0; // break from loop and return zero
                           }
                          break;
                        // everything else must be a REAL error. The question is which.
                        // After the 'WSAWOULDBLOCK'-thingy, the next error we often got
                        // was iWSAErrorCode == 10053 == WSAECONNABORTED .
                        case WSAECONNABORTED :
                          // The windoze documentation had this to say:
                          // > Software caused connection abort.
                          // > An established connection was aborted by the software
                          // > in your host computer, possibly due to
                          // > a data transmission time-out or protocol error.
                          // > The application should close the socket as it is no longer usable.
                          // In fact, got here when trying to send Ogg/Vorbis
                          //  to   67.207.139.49:8901  ,
                          //  and Paul's server replied with TCP 'RST' flag:
                          // > Some host TCP stacks may implement a half-duplex close sequence,
                          // > as Linux or HP-UX do. If such a host
                          // > actively closes a connection but still has not
                          // > read all the incoming data the stack already received
                          // > from the link, this host sends a RST instead of a FIN.
                          INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, iWSAErrorCode,
                              "Error in send: Connection aborted." );
                          iErrorCodeOrNumBytesWritten = -INET_ERROR_CONNECTION_ABORT;
                          nLoops = 0;
                          INET_CloseClient( pInetClient, "after 'WSAECONNABORTED'"/*pszWhy*/ );
                          pInetClient->fConnectionLost = TRUE;  // flag for the application
                          break;

                        case WSAECONNRESET :
                          // > An existing connection was forcibly closed by the remote host.
                          INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, iWSAErrorCode,
                              "Error in send: Connection reset (by remote host). Closing socket." );
                          iErrorCodeOrNumBytesWritten = -INET_ERROR_CONNECTION_RESET;
                          nLoops = 0;
                          INET_CloseClient( pInetClient, "after 'WSAECONNRESET'"/*pszWhy*/ );
                          pInetClient->fConnectionLost = TRUE;  // flag for the application
                          break;

                        default :
                          // Some other kind of most likely 'unrecoverable' error :
                          INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_ERROR|INET_EVENT_FLAG_SHOW_TIMESTAMP, iWSAErrorCode,
                              "Error from send: %s . Closing socket.",
                              INET_WinsockErrorCodeToString(iWSAErrorCode) );
                          iErrorCodeOrNumBytesWritten = -INET_ERROR_CONNECTION_ABORT;
                          nLoops = 0;
                          INET_CloseClient( pInetClient, "after some error"/*pszWhy*/ );
                          pInetClient->fConnectionLost = TRUE;  // flag for the application
                          break;
                     } // end switch( iWSAErrorCode )
                  } // end else <iErrorCodeOrNumBytesWritten<=0>
                 if( nMillisecondsWaited>0 && nMillisecondsWaited>=iTimeout_ms )
                  { break;
                  }
               }
              break; // end case INET_PROTOCOL_HTTP
           default:  // any other protocol isn't supported here yet !
              break;
         } // end switch( pInetClient->iProtocol )
      } // end if( pInetClient->is_internet_resource )
   } // end if < valid source >

  if( nMillisecondsWaited > pInetClient->nMillisecondsWaited )
   { pInetClient->nMillisecondsWaited = nMillisecondsWaited;
   }
  else
   { if(  pInetClient->nMillisecondsWaited>0 )
      { --pInetClient->nMillisecondsWaited;
      }
   }
  if( nMillisecondsWaited > pInetClient->nMillisecondsWaited_peak )
   { pInetClient->nMillisecondsWaited_peak = nMillisecondsWaited;
   }


  return iErrorCodeOrNumBytesWritten;
} // end INET_SendImmediately()


/****************************************************************************/
CPROT int INET_GetTxBufferUsage_kByte( T_INET_Client *pInetClient )
{
  // Only return "zero" if the transmit-buffer is COMPLETELY empty:
  return (pInetClient->nBytesInTxBuffer+1023) / 1024;
}


/****************************************************************************/

CPROT BOOL INET_AppendDataToTxBuffer(  // does what the name says...
     T_INET_Client *pInetClient,    // [in,out] web client data
     BYTE *pbSource, int iLength)   // [in] source data, and number of bytes to send
{
  if( (pbSource!=NULL) && (iLength>0) )  // is there really something to send ?
   {
     // If there's already 'something' in the buffer, but enough space left (it should),
     //  then append the data to the buffer contents to minimize the number of
     //  send()-calls .
     // So: Append the 'new' data to the buffer if possible :
     if( (pInetClient->nBytesInTxBuffer + iLength) < (long)sizeof(pInetClient->cTxBuffer) )
      { // the 'new' data fit inside the buffer, so append them to those data
        // which still wait for transmission:
        memmove( pInetClient->cTxBuffer + pInetClient->nBytesInTxBuffer/*dest*/,
                 pbSource, iLength );
        pInetClient->nBytesInTxBuffer += iLength;
        return TRUE;  // successfully appended the data to the transmit-buffer
      }
   } // end if < valid source >
  return FALSE;
} // end INET_AppendDataToTxBuffer()


/****************************************************************************/
CPROT int INET_Write(  // "continues" writing/sending if necessary
     T_INET_Client *pInetClient,    // [in,out] web client data
     int  iTimeout_ms,              // [in] max. timeout in milliseconds (0=non-blocking)
     int  iOptions,                 // [in] combination of INET_WR_OPTION_....
     BYTE *pbSource, int iLength)   // [in] source data, and number of bytes to send
  // To minimize the number of calls of send() ,
  //    the caller may specify the following flags in iOptions :
  //    INET_WR_OPTION_DONT_FLUSH : "Dont try to send already,
  //                                 I know there's more to send in the next call".
  //    INET_WR_OPTION_FLUSH_NOW  : "Send any data waiting in the buffer" .
  //    If none of the flags is set, the data-to-be may(!) be sent without buffering,
  //    especially when the block is 'quite long'( iLength > ? kByte) .
  //
  // Return value : number of bytes actually written, or a NEGATIVE error code
  //        compatible with c:\cbproj\SoundUtl\ErrorCodes.cpp::ErrorCodeToString()
{
  int iErrorCodeOrNumBytesWritten = 0;
  int nErrorCodeOrBytesSent2;
  // int n, nLoops = 2;

  if( pInetClient->sock==INVALID_SOCKET )
   { return -INET_ERROR_SOCKET_NOT_OPEN;
   }


  if( (pbSource!=NULL) && (iLength>0) )  // is there really something to send ?
   {
     // If there's already 'something' in the buffer, but enough space left (it should),
     //  then append the data to the buffer contents to minimize the number of
     //  send()-calls .
     // So: Append the 'new' data to the buffer if possible :
     if( INET_AppendDataToTxBuffer( pInetClient, pbSource, iLength ) )
      { // success .. added all "new" data BEFORE trying to send()
        iErrorCodeOrNumBytesWritten = iLength;
        iLength = 0;  // nothing more to append later
      }
   } // end if < valid source >


  // If the caller says 'INET_WR_OPTION_DONT_FLUSH' don't try to send yet (he'll have a reason for it).
  if( iOptions & INET_WR_OPTION_DONT_FLUSH )
   { // don't flush; new data already appended to the buffer; so we're through !
     return iErrorCodeOrNumBytesWritten;
   }

  // Arrived here: It's ok to flush the buffer if there's something in it..
  if( pInetClient->nBytesInTxBuffer > 0 ) // ok, there's something to send !
   {
     nErrorCodeOrBytesSent2 = INET_SendImmediately(  // "continues" writing/sending if necessary
                  pInetClient, iTimeout_ms, iOptions,
                  (BYTE*)pInetClient->cTxBuffer, pInetClient->nBytesInTxBuffer );
     if( nErrorCodeOrBytesSent2 < 0 )
      { // An ERROR in INET_SendImmediately() -> return "without success"
        return nErrorCodeOrBytesSent2;
      }
     else // nErrorCodeOrBytesSent2 >= 0 : Sent from the buffer, but was it 'all' ?
      { pInetClient->nBytesInTxBuffer -= nErrorCodeOrBytesSent2;
        if( pInetClient->nBytesInTxBuffer > 0 )
         { // oops.. the damned buffer couldn't be flushed COMPLETELY !
           if( nErrorCodeOrBytesSent2 > 0 ) // But at least A PART was sent ...
            { // Unlikely to happen, BUT (from the specification of send(<len>) :
              // > If no error occurs, send returns the total number of bytes sent.
              // > Note that this can be less than the number indicated by len.
              // So remove 'successfully sent bytes' at the start of the buffer
              // and shift the remaining bytes up.
              // BTW Did you know that memcpy() may fail if souce+dest overlap ?
              // Thus: use memmove here, not memcpy !
              memmove( pInetClient->cTxBuffer/*dest*/,
                 pInetClient->cTxBuffer+nErrorCodeOrBytesSent2/*src*/,
                 pInetClient->nBytesInTxBuffer); /* number of bytes to move*/
            }
           else // output buffer still not empty, and send() sent NOTHING AT ALL:
            { // In this case, do NOT call memmove() - it may crash with zero bytes to copy
            } // end else < nErrorCodeOrBytesSent2 == 0 >
         }
        else // Good news: successfully flushed the buffer completely.
         { pInetClient->nBytesInTxBuffer = 0;
         }
      } // end else < nErrorCodeOrBytesSent2 >= 0 >
   } // end if( pInetClient->nBytesInTxBuffer > 0 )


  if( (pbSource!=NULL) && (iLength>0) ) // STILL something remaining which could NOT be sent yet ?
   { // (because the buffer was already too full on entry into this function)
     if( INET_AppendDataToTxBuffer( pInetClient, pbSource, iLength ) )
      { iErrorCodeOrNumBytesWritten = iLength;
      }
   } // end if < valid source >

  return iErrorCodeOrNumBytesWritten;
} // end INET_Write()



/****************************************************************************/
void INET_CloseClient(
        T_INET_Client *pInetClient, // [in,out] web client data
        char * pszWhy) // [in, optional] reason for closing, may be NULL.
                       // Added for the display on the 'Analyse / play audio stream' panel,
                       // see SpecMain.cpp : TSpectrumLab::OpenInputFileAnalysis(),
                       // where we unfortunately need to connect the same server TWICE,
                       // e.g. pszWhy = "after analysing the format" .
{
  char sz255Msg[256];
  if( pInetClient != NULL )
   { if( pInetClient->sock!=INVALID_SOCKET )
      {
        strcpy( sz255Msg, "Closing socket connection" );
        if( pszWhy != NULL )
         { strcat( sz255Msg, " " );
           strncat( sz255Msg, pszWhy, 200 );
         }
        INET_InvokeCallback( pInetClient, INET_EVENT_TYPE_INFO|INET_EVENT_FLAG_SHOW_TIMESTAMP,
                             0/*progress or event*/, sz255Msg );
        // ex: close(sock); // NOT FOR WINDOZE ! A "SOCKET" is a not "FILE HANDLE" !
        // A socket is a socket, and a file is a file. They are DIFFERENT WORLDS !
        // From the Borland C++Builder help files:
        // > When a session has been completed, a closesocket must be performed.
        // 2015-02-01 : To test the behaviour of THE APPLICATION (SpecLab)
        //   when a socket breaks down, simulate a WLAN "breakdown" with the
        //   rarely used 'Radio' sliding switch on the frontside of the Lenovo Z61m :
        //   First, the error message "Connection reset by peer" showed up,
        //     followed by "Closing socket connection" a few seconds later,
        //     when SL called INET_CloseClient() through ..PlayNextFileFromList()
        //      -> C_AnyAudioFileIO::CloseFile() -> C_VorbisFileIO::CloseFile()
        //          -> INET_CloseClient() .
        closesocket( pInetClient->sock );
        pInetClient->sock = INVALID_SOCKET; // don't try to close a 2nd time
      }
     pInetClient->nBytesInRxBuffer = pInetClient->nBytesInTxBuffer = 0;
   }
} // end INET_CloseClient()


//---------------------------------------------------------------------------
// Various code snippets added to "play" with WebSockets .. 2019-01 :
//---------------------------------------------------------------------------

static const BYTE base64_table[65] =
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

//---------------------------------------------------------------------------
int INET_EncodeBase64( BYTE *pbSrc, int iSrcLen,
                       BYTE *pbDst, int iMaxDstLen,
                       int  iMaxCharsPerLine ) // [in] for optional "limited line lengths". 0 = all in ONE line of text.
  // Base64 encode
  // [in]  pbSrc[0..iSrcLen-1]    : Data (a block of bytes) to be encoded
  // [out] pbDst[0..iMaxDstLen-1] : Encoded output (base64), array with limited capacity !
  //           To be safe, use a destination with at least
  //           < iSrcLen * 4 / 3 + 4 + iSrcLen / 70 > bytes.
  //           If iMaxDstLen permits, pbDst will be terminated with 0x00.
  // Returns the length of the ENCODED "payload",  measured in bytes,
  // including '\n' characters for the line breaks, and a trailing zero-byte.
  // Returns the number of bytes (and thus CHARACTERS) placed in the output buffer,
  //  or ZERO if the destination block is too short.
  // UNLIKE THE ORIGINAL IMPLEMENTATION (found somewhere on the web years ago),
  //        this one does not append a stupid '\n' character if the output
  //        fits in a single line, because in almost any case, the CALLER
  //        had to remove that stupid character again (e.g. WebSocket) .
{
  BYTE *pbDstPos;
  BYTE *pbSrcEndstop, *pbSrcPos;
  int out_len, line_len;

  out_len = iSrcLen * 4 / 3 + 4; /* 3-byte blocks to 4-byte */
  if( iMaxCharsPerLine > 0 )
   { out_len += out_len / iMaxCharsPerLine; /* line feeds */
   }
  out_len++; // add one byte for the nul termination (to treat the result as a "C"-string)
  if( out_len > iMaxDstLen )
   { return 0; // > returns ZERO if the destination block is too short
   }

  pbSrcEndstop = pbSrc + iSrcLen;
  pbSrcPos = pbSrc;
  pbDstPos = pbDst;
  line_len = 0;
  while( (pbSrcEndstop - pbSrcPos) >= 3)
   { *pbDstPos++ = base64_table[pbSrcPos[0] >> 2];
     *pbDstPos++ = base64_table[((pbSrcPos[0] & 0x03) << 4) | (pbSrcPos[1] >> 4)];
     *pbDstPos++ = base64_table[((pbSrcPos[1] & 0x0f) << 2) | (pbSrcPos[2] >> 6)];
     *pbDstPos++ = base64_table[  pbSrcPos[2] & 0x3f];
     pbSrcPos += 3;
     line_len += 4;
     if( iMaxCharsPerLine > 0 ) // e.g. 72 ? Or 76 for MIME ? Nnngrrr !
      { if (line_len >= 72) // fortunately, the PARSER is smart enough to skip any '\n'
         {
           *pbDstPos++ = '\n';
           line_len = 0;
         }
      }
   }

  if( pbSrcPos < pbSrcEndstop ) // some remaining input, less than 3 bytes ..
   {
     *pbDstPos++ = base64_table[pbSrcPos[0] >> 2];
     if( (pbSrcEndstop - pbSrcPos) == 1)
      { *pbDstPos++ = base64_table[(pbSrcPos[0] & 0x03) << 4];
        *pbDstPos++ = '=';
      }
     else
      { *pbDstPos++ = base64_table[((pbSrcPos[0] & 0x03) << 4)
                                  | (pbSrcPos[1] >> 4)];
        *pbDstPos++ = base64_table[ (pbSrcPos[1] & 0x0f) << 2];
      }
     *pbDstPos++ = '=';
     line_len += 4;
   }

  if( iMaxCharsPerLine > 0 )
   { if (line_len) // for "nice-looking" output, line break after the last fragment
      { *pbDstPos++ = '\n';
      }
   }

  *pbDstPos = '\0';
  return pbDstPos - pbDst;
} // end INET_EncodeBase64()


//---------------------------------------------------------------------------
int INET_DecodeBase64( BYTE *pbSrc, int iSrcLen,
                       BYTE *pbDst, int iMaxDstLen )
  // Base64 decoder. Not the fastest, not elegant, but not as bulky as others,
  //                 and not poisioned by anything that isn't plain old "C".
  // [in]  pbSrc[0..iSrcLen-1]    : Data (base64) to be decoded
  // [out] pbDst[0..iMaxDstLen-1] : Decoded output, array with limited capacity !
  //           If iMaxDstLen permits, pbDst will be terminated with 0x00.
  // Returns the length of the DECODED "payload", measured in bytes.
  // Unlike many implementations floating around the planet,
  // this one does NOT use dynamically allocated memory blocks.
  // It's the caller's decision to use whatever he wants to.
{
  BYTE dtable[256], *pbDstPos, block[4], tmp;
  int  i, count; // , out_len;  // who needs a "size_t" ? We don't.
  int pad = 0;
  int nCharsDecoded;

  memset(dtable, 0x80, sizeof(dtable) );
  for (i = 0; i < sizeof(base64_table) - 1; i++)
   { dtable[base64_table[i]] = (unsigned char) i;
   }
  dtable['='] = 0;

  count = 0;
  for (i = 0; i < iSrcLen; i++)
   { if (dtable[pbSrc[i]] != 0x80)
      { count++;
      }
   }

  if (count == 0 || count % 4)
   { return 0;  // something wrong with the base64-encoded input (not N*4 "valid" characters)
   }

  // out_len = count / 4 * 3;
  pbDstPos = pbDst;

  count = 0;
  for (i = 0; i < iSrcLen; i++)
   {
     tmp = dtable[pbSrc[i]];
     if (tmp == 0x80)  // not a "valid" Base64 char ? Skip it, including '\n'
      { continue;
      }

     if (pbSrc[i] == '=')
      { pad++;
      }
     block[count] = tmp;
     count++;
     if (count == 4)
      { *pbDstPos++ = (block[0] << 2) | (block[1] >> 4);
        *pbDstPos++ = (block[1] << 4) | (block[2] >> 2);
        *pbDstPos++ = (block[2] << 6) | block[3];
        count = 0;
        switch(pad)
         { case 0 : // not padding right now
              break;
           case 1 :
              pbDstPos--;
              break;
           case 2 :
              pbDstPos -= 2;
              break;
           default: // invalid padding
              return 0;
         }
      } // end if( count==4 )
   }   // end for i ..

  nCharsDecoded = pbDstPos - pbDst;
  if( nCharsDecoded < iMaxDstLen ) // if possible, provide a trailing zero...
   { pbDst[nCharsDecoded] = 0x00;
   }

  return nCharsDecoded; // returns the length of the DECODED payload in bytes
} // end INET_DecodeBase64()



//---------------------------------------------------------------------------

// SHA-1 in C. By Steve Reid <steve@edmweb.com>.  100% Public Domain.

//    Not such a "Secure Hash Algorithm" as expected, but even "WebSockets"

//   (as specified in 2013) and TLS still used it in 2019.

//---------------------------------------------------------------------------


  // Test Vectors (from FIPS PUB 180-1)
  // "abc"
  //   A9993E36 4706816A BA3E2571 7850C26C 9CD0D89D
  // "abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq"
  //   84983E44 1C3BD26E BAAE4AA1 F95129E5 E54670F1
  // A million repetitions of "a"
  //   34AA973C D4C4DAA4 F61EEB2B DBAD2731 6534016F
  // From https://de.wikipedia.org/wiki/Secure_Hash_Algorithm :
  //   "Franz jagt im komplett verwahrlosten Taxi quer durch Bayern"
  //   68ac9064 95480a34 04beee48 74ed853a 037a7a8f

#define LITTLE_ENDIAN /* This should be #define'd if true.   */
#define SHA1HANDSOFF  /* Copies data before messing with it. */

void SHA1Transform(unsigned long state[5], unsigned char buffer[64]);
void INET_SHA1_Init(  T_INET_SHA1_Context* context);
void INET_SHA1_Update(T_INET_SHA1_Context* context, unsigned char* data, unsigned int len);
void INET_SHA1_Final( T_INET_SHA1_Context* context, unsigned char digest[20]);

#define rol(value, bits) (((value) << (bits)) | ((value) >> (32 - (bits))))

/* blk0() and blk() perform the initial expand. */
/* I got the idea of expanding during the round function from SSLeay */
#ifdef LITTLE_ENDIAN
#define blk0(i) (block->l[i] = (rol(block->l[i],24)&0xFF00FF00) \
    |(rol(block->l[i],8)&0x00FF00FF))
#else
#define blk0(i) block->l[i]
#endif
#define blk(i) (block->l[i&15] = rol(block->l[(i+13)&15]^block->l[(i+8)&15] \
    ^block->l[(i+2)&15]^block->l[i&15],1))

/* (R0+R1), R2, R3, R4 are the different operations used in SHA1 */
#define R0(v,w,x,y,z,i) z+=((w&(x^y))^y)+blk0(i)+0x5A827999+rol(v,5);w=rol(w,30);
#define R1(v,w,x,y,z,i) z+=((w&(x^y))^y)+blk(i)+0x5A827999+rol(v,5);w=rol(w,30);
#define R2(v,w,x,y,z,i) z+=(w^x^y)+blk(i)+0x6ED9EBA1+rol(v,5);w=rol(w,30);
#define R3(v,w,x,y,z,i) z+=(((w|x)&y)|(w&x))+blk(i)+0x8F1BBCDC+rol(v,5);w=rol(w,30);
#define R4(v,w,x,y,z,i) z+=(w^x^y)+blk(i)+0xCA62C1D6+rol(v,5);w=rol(w,30);


/* Hash a single 512-bit block. This is the core of the algorithm. */

void SHA1Transform(unsigned long state[5], unsigned char buffer[64])
{
  unsigned long a, b, c, d, e;
  typedef union {
    unsigned char c[64];
    unsigned long l[16];
  } CHAR64LONG16;
  CHAR64LONG16* block;
#ifdef SHA1HANDSOFF
  static unsigned char workspace[64];
    block = (CHAR64LONG16*)workspace;
    memcpy(block, buffer, 64);
#else
    block = (CHAR64LONG16*)buffer;
#endif
    /* Copy context->state[] to working vars */
    a = state[0];
    b = state[1];
    c = state[2];
    d = state[3];
    e = state[4];
    /* 4 rounds of 20 operations each. Loop unrolled. */
    R0(a,b,c,d,e, 0); R0(e,a,b,c,d, 1); R0(d,e,a,b,c, 2); R0(c,d,e,a,b, 3);
    R0(b,c,d,e,a, 4); R0(a,b,c,d,e, 5); R0(e,a,b,c,d, 6); R0(d,e,a,b,c, 7);
    R0(c,d,e,a,b, 8); R0(b,c,d,e,a, 9); R0(a,b,c,d,e,10); R0(e,a,b,c,d,11);
    R0(d,e,a,b,c,12); R0(c,d,e,a,b,13); R0(b,c,d,e,a,14); R0(a,b,c,d,e,15);
    R1(e,a,b,c,d,16); R1(d,e,a,b,c,17); R1(c,d,e,a,b,18); R1(b,c,d,e,a,19);
    R2(a,b,c,d,e,20); R2(e,a,b,c,d,21); R2(d,e,a,b,c,22); R2(c,d,e,a,b,23);
    R2(b,c,d,e,a,24); R2(a,b,c,d,e,25); R2(e,a,b,c,d,26); R2(d,e,a,b,c,27);
    R2(c,d,e,a,b,28); R2(b,c,d,e,a,29); R2(a,b,c,d,e,30); R2(e,a,b,c,d,31);
    R2(d,e,a,b,c,32); R2(c,d,e,a,b,33); R2(b,c,d,e,a,34); R2(a,b,c,d,e,35);
    R2(e,a,b,c,d,36); R2(d,e,a,b,c,37); R2(c,d,e,a,b,38); R2(b,c,d,e,a,39);
    R3(a,b,c,d,e,40); R3(e,a,b,c,d,41); R3(d,e,a,b,c,42); R3(c,d,e,a,b,43);
    R3(b,c,d,e,a,44); R3(a,b,c,d,e,45); R3(e,a,b,c,d,46); R3(d,e,a,b,c,47);
    R3(c,d,e,a,b,48); R3(b,c,d,e,a,49); R3(a,b,c,d,e,50); R3(e,a,b,c,d,51);
    R3(d,e,a,b,c,52); R3(c,d,e,a,b,53); R3(b,c,d,e,a,54); R3(a,b,c,d,e,55);
    R3(e,a,b,c,d,56); R3(d,e,a,b,c,57); R3(c,d,e,a,b,58); R3(b,c,d,e,a,59);
    R4(a,b,c,d,e,60); R4(e,a,b,c,d,61); R4(d,e,a,b,c,62); R4(c,d,e,a,b,63);
    R4(b,c,d,e,a,64); R4(a,b,c,d,e,65); R4(e,a,b,c,d,66); R4(d,e,a,b,c,67);
    R4(c,d,e,a,b,68); R4(b,c,d,e,a,69); R4(a,b,c,d,e,70); R4(e,a,b,c,d,71);
    R4(d,e,a,b,c,72); R4(c,d,e,a,b,73); R4(b,c,d,e,a,74); R4(a,b,c,d,e,75);
    R4(e,a,b,c,d,76); R4(d,e,a,b,c,77); R4(c,d,e,a,b,78); R4(b,c,d,e,a,79);
    /* Add the working vars back into context.state[] */
    state[0] += a;
    state[1] += b;
    state[2] += c;
    state[3] += d;
    state[4] += e;
    /* Wipe variables */ // (but modern compilers tell you this is "never used"..)
    a = b = c = d = e = 0;
    (void)a;   // shut up Borland, we don't want to leave traces on the stack !
}


/* INET_SHA1_Init - Initialize new context */

void INET_SHA1_Init(T_INET_SHA1_Context* context)
{
    /* SHA1 initialization constants */
    context->state[0] = 0x67452301;
    context->state[1] = 0xEFCDAB89;
    context->state[2] = 0x98BADCFE;
    context->state[3] = 0x10325476;
    context->state[4] = 0xC3D2E1F0;
    context->count[0] = context->count[1] = 0;
}


/* Run your data through this. */

void INET_SHA1_Update(T_INET_SHA1_Context* context, unsigned char* data, unsigned int len)
{
  unsigned int i, j;

  j = (context->count[0] >> 3) & 63;
  if ((context->count[0] += len << 3) < (len << 3)) context->count[1]++;
  context->count[1] += (len >> 29);
  if ((j + len) > 63)
   { memcpy(&context->buffer[j], data, (i = 64-j));
     SHA1Transform(context->state, context->buffer);
     for ( ; i + 63 < len; i += 64)
      { SHA1Transform(context->state, &data[i]);
      }
     j = 0;
   }
  else
   { i = 0;
   }
  memcpy(&context->buffer[j], &data[i], len - i);
} // end INET_SHA1_Update()


/* Add padding and return the message digest. */

void INET_SHA1_Final(T_INET_SHA1_Context* context, unsigned char digest[20])
{
  unsigned long i, j;
  unsigned char finalcount[8];

  for (i = 0; i < 8; i++)
   { finalcount[i] = (unsigned char)((context->count[(i >= 4 ? 0 : 1)]
                      >> ((3-(i & 3)) * 8) ) & 255);  /* Endian independent */
   }
  INET_SHA1_Update(context, (unsigned char *)"\200", 1);
  while ((context->count[0] & 504) != 448)
   { INET_SHA1_Update(context, (unsigned char *)"\0", 1);
   }
  INET_SHA1_Update(context, finalcount, 8);  /* Should cause a SHA1Transform() */
  for (i = 0; i < 20; i++)
   { digest[i] = (unsigned char)
         ((context->state[i>>2] >> ((3-(i & 3)) * 8) ) & 255);
   }
  /* Wipe variables */
  i = j = 0;
  (void)i; // defeat warning from Borland C : "blah was assigned a value that is never used".
           // Guess Steve's intention for 'wiping variables' is NOT to leave a trace
           // of what has been calculated on the stack .
  (void)j; // again: shut up, Borland (blubb assigned a value that is never used)
  memset(context->buffer, 0, 64);
  memset(context->state, 0, 20);
  memset(context->count, 0, 8);
  memset(&finalcount, 0, 8);
#ifdef SHA1HANDSOFF  /* make SHA1Transform overwrite it's own static vars */
  SHA1Transform(context->state, context->buffer);
#endif
} // end INET_SHA1_Final()


//---------------------------------------------------------------------------
BOOL INET_WebSecSocketKeyToWebSocketAcceptValue(
         char* pszSecWebSocketKey,       // [in]  value from field "Sec-Websocket-Key"
         char* cpDestSecWebSocketAccept, // [out] value for "Sec-Websocket-Accept"
         char* cpDestEndstop )           // [in] endstop for the destination (max. position of the string's trailing zero)
{ char sz255[256];
  T_INET_SHA1_Context sha1_context; // temporary instance to calculate SHA1..
  BYTE b20Digest[20];  // .. as "digest" to generate the "Sec-WebSocket-Accept"-value
  int  iEncodedLength;
  SL_strncpy( sz255, pszSecWebSocketKey, sizeof(sz255)-40/*max dest length*/ );
  SL_strncat( sz255, "258EAFA5-E914-47DA-95CA-C5AB0DC85B11", sizeof(sz255)-1 );
  INET_SHA1_Init(&sha1_context);  // Prepare SHA-1 calculation
  INET_SHA1_Update(&sha1_context, (BYTE*)sz255, SL_strnlen(sz255,255) ); // add to SHA-1 hash ("digest")
  INET_SHA1_Final(&sha1_context, b20Digest ); // Finish SHA-1 calculation
  iEncodedLength = INET_EncodeBase64( b20Digest/*pbSrc*/, 20/*iSrcLen*/,
                      (BYTE*)cpDestSecWebSocketAccept/*pbDst*/,
                      cpDestEndstop - cpDestSecWebSocketAccept/*iMaxDstLen*/,
                      0/*all in ONE line*/ );
  if( iEncodedLength < 20*4/3 ) // oops.. this cannot be correct for a 20-byte "digest" !
   { return FALSE;
   }                    
  // Example : sz255="qZUNEqiPTUQhPTem1RFaUA==258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
  //       b20Digest= 4c e7 07 a1 70 45 cf 13 e2 15 af 33 18 b3 e9 99 23 8f 82 6c
  //                  (ok, according to www.sha1generator.de/ )
  //       cpDestSecWebSocketAccept = "TOcHoXBFzxPiFa8zGLPpmSOPgmw="
  //                  (calculated with the "Hexadecimal -> base64 string encoder" at tomeko.net/online_tools/hex_to_base64.php)
  if( (cpDestSecWebSocketAccept+iEncodedLength) < cpDestEndstop )
   { cpDestSecWebSocketAccept[ iEncodedLength ] = '\0'; // ok (sufficient capacity)
     return TRUE;
   }
  else
   { *cpDestEndstop = '\0'; // error, truncated, this won't be a valid "Sec-WebSocket-Accept" - value
     return FALSE;
   }
} // end INET_WebSecSocketKeyToWebSocketAcceptValue()



#if(0)  // TEST application for the above SHA-1 implementation, by Steve Reid
        // Of course we don't want to have a "main" in this module !
        // A *very short* test for the SHA-1 is contained in INET_Init() .

int main(int argc, char** argv)
{
  int i, j;
  T_INET_SHA1_Context context;
  unsigned char digest[20], buffer[16384];
  FILE* file;

  if (argc > 2) {
        puts("Public domain SHA-1 implementation - by Steve Reid <steve@edmweb.com>");
        puts("Produces the SHA-1 hash of a file, or stdin if no file is specified.");
        exit(0);
    }
    if (argc < 2) {
        file = stdin;
    }
    else {
        if (!(file = fopen(argv[1], "rb"))) {
            fputs("Unable to open file.", stderr);
            exit(-1);
        }
    }
    INET_SHA1_Init(&context);
    while (!feof(file)) {  /* note: what if ferror(file) */
        i = fread(buffer, 1, 16384, file);
        INET_SHA1_Update(&context, buffer, i);
    }
    INET_SHA1_Final(digest, &context);
    fclose(file);
    for (i = 0; i < 5; i++) {
        for (j = 0; j < 4; j++) {
            printf("%02X", digest[i*4+j]);
        }
        putchar(' ');
    }
    putchar('\n');
    exit(0);
}

#endif // (0)

//---------------------------------------------------------------------------
int INET_Init(void) // -> return value : >= 0 when ok, < 0 indicates an error.
  // Returns a NEGATIVE error code if any of the built-in self tests failed.
  // Besides self-testing a few function implemented in this module,
  // doesn't do anything important.
{
  BYTE b256[256], b20Digest[20];
  char sz255[256];
  int  i, iEncodedLength, iDecodedLength;
  T_INET_SHA1_Context sha1_context;


  // Perform a few(!) internal self tests, save headaches later when stuff gets REALLY complex !
  iEncodedLength = INET_EncodeBase64( (BYTE*)szFranz, strlen(szFranz), b256, sizeof(b256), 0 );
  if( iEncodedLength != (int)strlen( szFranzBase64 ) )  // 2019-01-21 : 80 bytes
   { return -1;  // not the expected length from the Base64 encoder
   }
  if( memcmp( b256, szFranzBase64, iEncodedLength) != 0 ) // 2019-01-21: "RnJh.."
   { return -2;  // not the expected data from the Base64 encoder
   }
  iDecodedLength = INET_DecodeBase64( (BYTE*)szFranzBase64, strlen(szFranzBase64), b256, sizeof(b256) );
  if( iDecodedLength != (int)strlen( szFranz ) )
   { return -3;  // not the expected length from the Base64 decoder
   }
  if( memcmp( b256, szFranz, iDecodedLength) != 0 )
   { return -4;  // not the expected data from the Base64 decoder
   }
  iEncodedLength = INET_EncodeBase64( (BYTE*)"a", 1, b256, sizeof(b256), 0/*no line-breaks*/ );
  if( iEncodedLength != 4 )
   { return -5;  // not the expected length from the Base64 decoder
   }
  if( memcmp( b256, "YQ==", 4) != 0 )
   { return -6;  // not the expected data from the Base64 encoder for a SINGLE character
   }
  iDecodedLength = INET_DecodeBase64( b256, iEncodedLength, (BYTE*)sz255,255 ); // decode back to "a" again ?
  if( iDecodedLength != 1 )
   { return -7;  // not the expected length from the Base64 decoder for a single byte
   }
  if( memcmp( sz255, "a", iDecodedLength) != 0 )
   { return -8;  // not the expected data from the Base64 decoder
   }
  iEncodedLength = INET_EncodeBase64( (BYTE*)"ab", 2, b256, sizeof(b256), 0/*no line-breaks*/ );
  if( iEncodedLength != 4 )
   { return -9;  // not the expected length from the Base64 decoder
   }
  if( memcmp( b256, "YWI=", 4) != 0 )
   { return -10;  // not the expected data from the Base64 encoder for a SINGLE character
   }
  iDecodedLength = INET_DecodeBase64( b256, iEncodedLength, (BYTE*)sz255,255 ); // decode back to "ab" again ?
  if( iDecodedLength != 2 )
   { return -11;  // not the expected length from the Base64 decoder for a two-byte string
   }
  if( memcmp( sz255, "ab", iDecodedLength) != 0 )
   { return -12;  // not the expected data from the Base64 decoder
   }
  iEncodedLength = INET_EncodeBase64( (BYTE*)"abc", 3, b256, sizeof(b256), 0/*no line-breaks*/ );
  if( iEncodedLength != 4 )
   { return -13;  // not the expected length from the Base64 decoder
   }
  if( memcmp( b256, "YWJj", 4) != 0 )
   { return -14;  // not the expected data from the Base64 encoder for a SINGLE character
   }
  iDecodedLength = INET_DecodeBase64( b256, iEncodedLength, (BYTE*)sz255,255 ); // decode back to "abc" again ?
  if( iDecodedLength != 3 )
   { return -15;  // not the expected length from the Base64 decoder for a two-byte string
   }
  if( memcmp( sz255, "abc", iDecodedLength) != 0 )
   { return -16;  // not the expected data from the Base64 decoder
   }

  INET_SHA1_Init(&sha1_context);  // Prepare SHA-1 calculation
  INET_SHA1_Update(&sha1_context, (BYTE*)szFranz, strlen(szFranz) ); // add to SHA-1 hash ("digest")
  INET_SHA1_Final(&sha1_context, b20Digest ); // Finish SHA-1 calculation
  for( i=0; i<20; ++i )   // 2019-01-21 : Test passed (all 20 bytes of the "digest")
   { if( b20Digest[i] != b20FranzSHA1[i] )
      { return -17;  // something screwed up with the SHA-1 calculation
      }
   }

  // Test "Sec-WebSocket-Key" -> "Sec-WebSocket-Accept",
  // using data from an OpenWebSDR sessession (with Chrome's "Network" tools)
  if( ! INET_WebSecSocketKeyToWebSocketAcceptValue( // generate value for "Sec-Websocket-Accept"
         "qZUNEqiPTUQhPTem1RFaUA==",  // [in]  value from field "Sec-Websocket-Key"
         sz255,      // [out] cpDestSecWebSocketAccept, value for "Sec-Websocket-Accept"
         sz255+255)) // [in] cpDestEndstop, endstop for the destination (max. position of the string's trailing zero)
   { return -18;  // INET_WebSecSocketKeyToWebSocketAcceptValue() noticed an error itself
   }
  if( strcmp(sz255, "TOcHoXBFzxPiFa8zGLPpmSOPgmw=") != 0 )
   { return -19;  // INET_WebSecSocketKeyToWebSocketAcceptValue() produced a wrong "Sec-WebSocket-Accept" value (string)
   }

  return 1;  // successfully performed all built-in module tests

} // end INET_Init()



/* EOF < YHF_Inet.c > */
